# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import json
import os
import subprocess

import pyudev

from probert.utils import (
    read_sys_block_size_bytes,
    sane_block_devices,
    )

log = logging.getLogger('probert.lvm')


def _lvm_report(cmd, report_key):
    """ [pvs --reportformat=json -o foo,bar] report_key='pv'
     {
         "report": [
             {
                 "pv": [
                     {"pv_name":"/dev/md0", "vg_name":"vg0",
                      "pv_fmt":"lvm2", "pv_attr":"a--",
                      "pv_size":"<9.99g", "pv_free":"<6.99g"},
                     {"pv_name":"/dev/md1", "vg_name":"vg0",
                      "pv_fmt":"lvm2", "pv_attr":"a--",
                      "pv_size":"<9.99g", "pv_free":"<9.99g"}
                 ]
             }
         ]
     }
    """
    def _flatten_list(data):
        return [y for x in data for y in x]

    try:
        result = subprocess.run(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.DEVNULL)
        output = result.stdout.decode('utf-8')
    except subprocess.CalledProcessError as e:
        log.error('Failed to probe LVM devices on system: %s', e)
        return []

    if not output:
        return []

    reports = {}
    try:
        reports = json.loads(output)
    except json.decoder.JSONDecodeError as e:
        log.error('Failed to load LVM json report: %s', e)
        return []

    return _flatten_list([report.get(report_key)
                          for report in reports.get('report', [])
                          if report_key in report])


def probe_pvs_report():
    return _lvm_report(['pvs', '--reportformat=json'], 'pv')


def probe_vgs_report():
    report_cmd = ['vgs', '--reportformat=json', '--units=B',
                  '-o', 'vg_name,pv_name,pv_uuid,vg_size']
    return _lvm_report(report_cmd, 'vg')


def probe_lvs_report():
    return _lvm_report(['lvs'], 'lv')


def lvmetad_running():
    return os.path.exists(os.environ.get('LVM_LVMETAD_PIDFILE',
                                         '/run/lvmetad.pid'))


def lvm_scan():
    for cmd in [['pvscan'], ['vgscan']]:
        if lvmetad_running():
            cmd.append('--cache')
        try:
            subprocess.run(cmd, stdout=subprocess.DEVNULL,
                           stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError as e:
            log.error('Failed lvm_scan command %s: %s', cmd, e)


def activate_volgroups():
    """
    Activate available volgroups and logical volumes within.
    # found
    % vgchange -ay
      1 logical volume(s) in volume group "vg1sdd" now active

    # none found (no output)
    % vgchange -ay
    """

    # vgchange handles syncing with udev by default
    # see man 8 vgchange and flag --noudevsync
    result = subprocess.run(['vgchange', '--activate=y'], check=False,
                            stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    if result.stdout:
        log.info(result.stdout)


def extract_lvm_partition(probe_data):
    lv_id = "%s/%s" % (probe_data['DM_VG_NAME'], probe_data['DM_LV_NAME'])
    return (
        lv_id, {'fullname': lv_id,
                'name': probe_data['DM_LV_NAME'],
                'volgroup': probe_data['DM_VG_NAME'],
                'size': "%sB" % read_sys_block_size_bytes(
                    probe_data['DEVNAME'])})


def extract_lvm_volgroup(vg_name, report_data):
    """
    [
        {"vg_name":"vg0", "pv_name":"/dev/md0",
         "pv_uuid":"p3oDow-dRHp-L8jq-t6gQ-67tv-B8B6-JWLKZP",
         "vg_size":"21449670656B"},
        {"vg_name":"vg0", "pv_name":"/dev/md1",
         "pv_uuid":"pRR5Zn-c4a9-teVZ-TFaU-yDxf-FSDo-cORcEq",
         "vg_size":"21449670656B"}
    ]
    """
    def _int(size_val):
        if size_val and size_val.endswith('B'):
            return int(size_val[:-1])
        return 0

    devices = set()
    size = None
    for report in report_data:
        if report['vg_name'] == vg_name:
            vg_size = report['vg_size']
            # set size to the largest size we find
            if vg_size:
                # unset, take current value
                if not size:
                    size = vg_size
                # on set but mismatched values, keep the larger
                elif size != vg_size:
                    if _int(vg_size) > _int(size):
                        size = vg_size
            devices.add(report.get('pv_name'))

    if size is None:
        size = '0B'

    return (vg_name, {'name': vg_name,
                      'devices': sorted(list(devices)),
                      'size': size})


def probe(context=None):
    """ Probing for LVM devices requires initiating a kernel level scan
        of block devices to look for physical volumes, volume groups and
        logical volumes.  Once detected, the prober will activate any
        volume groups detected.

        The prober will refresh the udev context which brings in addition
        information relating to LVM devices.

        This prober relies on udev detecting devices via the 'DM_UUID'
        field and for each of such devices, the prober records the
        logical volume.

        For each logical volume, the prober determines the hosting
        volume_group and records detailed information about the group
        including members.  The process is repeated to determine the
        underlying physical volumes that are used to construct a
        volume group.

        Care is taken to handle scenarios where physical volumes are
        not yet allocated to a volume group (such as a linear VG).

        On newer systems (Disco+) the lvm2 software stack provides
        a rich reporting data dump in JSON format.  On systems with
        older LVM2 stacks, the LVM probe may be incomplete.
    """
    # scan and activate lvm vgs/lvs
    lvm_scan()
    activate_volgroups()

    # ignore supplied context, we need to read udev after scan/vgchange
    context = pyudev.Context()

    lvols = {}
    vgroups = {}
    pvols = {}
    vg_report = probe_vgs_report()

    for device in sane_block_devices(context):
        if 'DM_UUID' in device and device['DM_UUID'].startswith('LVM'):
            (lv_id, new_lv) = extract_lvm_partition(device)
            if lv_id not in lvols:
                lvols[lv_id] = new_lv
            else:
                log.error('Found duplicate logical volume: %s', lv_id)
                continue

            vg_name = device['DM_VG_NAME']
            (vg_id, new_vg) = extract_lvm_volgroup(vg_name, vg_report)
            if vg_id not in vgroups:
                vgroups[vg_id] = new_vg
            else:
                log.error('Found duplicate volume group: %s', vg_id)
                continue

            if vg_id not in pvols:
                pvols[vg_id] = new_vg['devices']

    lvm = {}
    if lvols:
        lvm.update({'logical_volumes': lvols})
    if pvols:
        lvm.update({'physical_volumes': pvols})
    if vgroups:
        lvm.update({'volume_groups': vgroups})

    return lvm



# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import pyudev
import subprocess

from probert.utils import (
    read_sys_block_size_bytes,
    sane_block_devices,
    udev_get_attributes,
    )
from probert import (bcache, dasd, dmcrypt, filesystem, lvm, mount, multipath,
                     raid, zfs)

log = logging.getLogger('probert.storage')


class StorageInfo():
    ''' properties:
        .type = [disk, partition, etc.}
        .name = /dev/sda
        .size = 123012034 (bytes)
        .serial = abcdefghijkl
        .vendor = Innotec
        .model = SolidStateRocketDrive
        .devpath = /devices
        .is_virtual =
        .raw = {raw dictionary}
    '''
    def __init__(self, probe_data):
        [self.name] = probe_data
        self.raw = probe_data.get(self.name)

        self.type = self.raw['DEVTYPE']
        self.size = int(self.raw['attrs']['size'])

    def _get_hwvalues(self, keys):
        for key in keys:
            try:
                return self.raw[key]
            except KeyError:
                log.debug(
                    'Failed to get key {} from interface {}'.format(key,
                                                                    self.name))

        return None

    @property
    def vendor(self):
        ''' Some disks don't have ID_VENDOR_* instead the vendor
            is encoded in the model: SanDisk_A223JJ3J3 '''
        v = self._get_hwvalues(['ID_VENDOR_FROM_DATABASE', 'ID_VENDOR',
                                'ID_VENDOR_ID'])
        if v is None:
            v = self.model
            if v is not None:
                return v.split('_')[0]
        return v

    @property
    def model(self):
        return self._get_hwvalues(['ID_MODEL_FROM_DATABASE', 'ID_MODEL',
                                   'ID_MODEL_ID'])

    @property
    def serial(self):
        return self._get_hwvalues(['ID_SERIAL', 'ID_SERIAL_SHORT'])

    @property
    def devpath(self):
        return self._get_hwvalues(['DEVPATH'])

    @property
    def is_virtual(self):
        return self.devpath.startswith('/devices/virtual/')


def blockdev_probe(context=None):
    """ Non-class method for extracting relevant block
        devices from pyudev.Context().
    """
    def _extract_partition_table(devname):
        cmd = ['sfdisk', '--bytes', '--json', devname]
        try:
            result = subprocess.run(cmd, stdout=subprocess.PIPE,
                                    stderr=subprocess.DEVNULL)
            output = result.stdout.decode('utf-8')
        except subprocess.CalledProcessError as e:
            log.error('Failed to probe partition table on %s:%s', devname, e)
            return None
        if not output:
            return None
        ptable = {}
        try:
            ptable = json.loads(output)
        except json.decoder.JSONDecodeError:
            log.exception('Failed to load sfdisk json output:')
        return ptable

    if not context:
        context = pyudev.Context()

    blockdev = {}
    for device in sane_block_devices(context):
        if device['MAJOR'] not in ["1", "7"]:
            attrs = udev_get_attributes(device)
            # update the size attr as it may only be the number
            # of blocks rather than size in bytes.
            attrs['size'] = \
                str(read_sys_block_size_bytes(device['DEVNAME']))
            blockdev[device['DEVNAME']] = dict(device)
            blockdev[device['DEVNAME']].update({'attrs': attrs})
            # include partition table info if present
            ptable = _extract_partition_table(device['DEVNAME'])
            if ptable:
                blockdev[device['DEVNAME']].update(ptable)

    return blockdev


class Storage():
    """ The Storage class includes a map of storage types that
        probert knows how to extract required information needed
        for installation and use.  Each storage module included
        provides a probe method which will prepare and probe the
        environment for the specific type of storage devices.

        The result of each probe is collected into a dictionary
        which is collected in the class .results attribute.

        The probe is non-destructive and read-only; however a
        probe module may load additional modules if they are not
        present.
    """
    probe_map = {
        'bcache': bcache.probe,
        'blockdev': blockdev_probe,
        'dasd': dasd.probe,
        'dmcrypt': dmcrypt.probe,
        'filesystem': filesystem.probe,
        'lvm': lvm.probe,
        'mount': mount.probe,
        'multipath': multipath.probe,
        'raid': raid.probe,
        'zfs': zfs.probe
    }

    def __init__(self, results={}):
        self.results = results
        self.context = pyudev.Context()

    def _get_probe_types(self):
        return {ptype for ptype, pfunc in self.probe_map.items() if pfunc}

    def probe(self, probe_types=None):
        default_probes = self._get_probe_types()
        if not probe_types:
            to_probe = default_probes
        else:
            to_probe = probe_types.intersection(default_probes)

        if len(to_probe) == 0:
            not_avail = probe_types.difference(default_probes)
            print('Requsted probes not available: %s' % probe_types)
            print('Valid probe types: %s' % default_probes)
            print('Unavilable probe types: %s' % not_avail)
            return self.results

        probed_data = {}
        for ptype in to_probe:
            pfunc = self.probe_map[ptype]
            probed_data[ptype] = pfunc(context=self.context)

        self.results = probed_data
        return probed_data



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple
import logging
import operator
import os
import re
import subprocess
from functools import reduce


log = logging.getLogger('probert.zfs')
ZfsListEntry = namedtuple('ZfsListEntry',
                          ('name', 'used', 'avail', 'refer', 'mountpoint'))


def parse_zdb_output(data):
    """ Parse structured zdb output into a dictionary.

    hogshead:
        version: 5000
        name: 'hogshead'
        vdev_tree:
            type: 'root'
            id: 0
            guid: 12392392111803944759
            children[0]:
                type: 'raidz'
                ashift: 12
                children[0]:
                    type: 'disk'
                    id: 0
                    guid: 13921270083288950156
                    path: '/dev/disk/by-id/usb-ST4000VN_000-1H4168-0:0-part1'
                    whole_disk: 1
                    DTL: 140
                    create_txg: 4
                    com.delphix:vdev_zap_leaf: 231
                children[1]:
                    type: 'disk'
                    id: 1
                    guid: 2635788368927674810
                    path: '/dev/disk/by-id/usb-ST4000VN_000-1H4168-0:1-part1'
                    whole_disk: 1
                    DTL: 139
                    create_txg: 4
                    com.delphix:vdev_zap_leaf: 232
    """

    def get_from_dict(datadict, maplist):
        return reduce(operator.getitem, maplist, datadict)

    def set_in_dict(datadict, maplist, value):
        get_from_dict(datadict, maplist[:-1])[maplist[-1]] = value

    def parse_line_key_value(line):
        """ use ': ' token to split line into key, value pairs

        com.delphi:vdev_zap_top: 230
                               ^^
                                `- span() = (24, 26)
        key = 'com.delphi:vdev_zap_top'
        value = '230'
        """
        match = re.search(r': ', line)
        if match:
            tok_start, tok_end = match.span()
            key, value = (line[:tok_start], line[tok_end:])
        else:
            key, value = line.split(':')

        return (key.lstrip(), value.replace("'", ""))

    # for each line in zdb output, calculate the nested level
    # based on indentation. Add key/value pairs for each line
    # and generate a list of keys to calcaulate where in the root
    # dictionary to set the value.
    root = {}
    lvl_tok = 4
    prev_item = []
    for line in data.splitlines():
        current_level = int((len(line) - len(line.lstrip(' '))) / lvl_tok)
        prev_level = len(prev_item) - 1
        key, value = parse_line_key_value(line)
        # TODO: handle children[N] keyname an convert to list
        if current_level == 0:
            root[key] = {}
            prev_item = [(current_level, key)]
        else:
            new_item_path = [item[1]
                             for item in prev_item[0: current_level]] + [key]
            if value:
                set_in_dict(root, new_item_path, value)
            else:
                set_in_dict(root, new_item_path, {})
                # we've dropped down a level, replace prev level key w/new key
                if current_level == prev_level:
                    prev_item.pop()
                prev_item.append((current_level, key))

    return root


def zdb_asdict(data=None):
    """ Convert output from zdb into a dictionary"""
    if not data:
        cmd = ['zdb']
        # exported, altroot, and uncached pools need -e
        if not os.path.exists('/etc/zfs/zpool.cache'):
            cmd.append('-e')
        try:
            result = subprocess.run(cmd, stdout=subprocess.PIPE,
                                    stderr=subprocess.DEVNULL)
        except (subprocess.CalledProcessError, FileNotFoundError):
            return {}

        data = result.stdout.decode('utf-8')

    return parse_zdb_output(data)


def zfs_list_filesystems(raw_output=False):
    cmd = ['zfs', 'list', '-Hp', '-t', 'filesystem']
    try:
        result = subprocess.run(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        return []

    data = result.stdout.decode('utf-8')
    if raw_output:
        return data

    # NAME, USED, AVAIL, REFER, MOUNTPOINT
    zfs_entries = []
    for line in data.splitlines():
        (name, used, avail, refer, mpoint) = line.split('\t')
        if mpoint == 'none':
            mpoint = None
        zfs_entries.append(ZfsListEntry(name, used, avail, refer, mpoint))

    return zfs_entries


def zfs_get_properties(zfs_name, raw_output=False):
    if not zfs_name:
        raise ValueError('Invalid zfs_name parameter: "%s"', zfs_name)

    cmd = ['zfs', 'get', 'all', '-Hp', zfs_name]
    try:
        result = subprocess.run(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.DEVNULL)
    except subprocess.ProcessExecutionError:
        return []

    data = result.stdout.decode('utf-8')
    if raw_output:
        return data

    # NAME, PROPERTY, VALUE, SOURCE
    zprops = {}
    for line in data.splitlines():
        (name, prop, value, source) = line.split('\t')
        zprops[prop] = {'value': value, 'source': source}

    return {zfs_name: {'properties': zprops}}


def is_zfs_device(device):
    return device.get('ID_FS_TYPE') == 'zfs_member'


def probe(context=None):
    """The ZFS prober examines the ZFS Dubugger (zdb) output which
    produces psuedo-json output.  This is converted to a dictionary
    where for each zpool, we can extract the datasets and determine
    which vdevs (linux block devices) are used to construct the the
    zpool.

    For each zpool and dataset, the prober further extracts all of
    the pool and filesystem (dataset) properties and captures if
    the values and whether they are local changes or defaults.

    The resulting output includes the converted zdb dump and
    a tree of datasets and their properties.
    """
    zdb = zdb_asdict()
    zpools = {}
    for zpool, zdb_dump in zdb.items():
        datasets = {}
        zlf = zfs_list_filesystems()
        for zfs_entry in zlf:
            datasets.update(zfs_get_properties(zfs_entry.name))
        zpools[zpool] = {'zdb': zdb_dump, 'datasets': datasets}

    return {'zpools': zpools}



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import imp
import importlib
import mock
import random
import string


def builtin_module_name():
    options = ('builtins', '__builtin__')
    for name in options:
        try:
            imp.find_module(name)
        except ImportError:
            continue
        else:
            print('importing and returning: %s' % name)
            importlib.import_module(name)
            return name


@contextlib.contextmanager
def simple_mocked_open(content=None):
    if not content:
        content = ''
    m_open = mock.mock_open(read_data=content)
    mod_name = builtin_module_name()
    m_patch = '{}.open'.format(mod_name)
    with mock.patch(m_patch, m_open, create=True):
        yield m_open


def random_string(length=8):
    """ return a random lowercase string with default length of 8"""
    return ''.join(
        random.choice(string.ascii_lowercase) for _ in range(length))



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple
import logging
import subprocess

MPath = namedtuple("MPath", ('device', 'serial', 'multipath', 'host_wwnn',
                             'target_wwnn', 'host_wwpn', 'target_wwpn',
                             'host_adapter'))
MMap = namedtuple("MMap", ('multipath', 'sysfs', 'paths'))
MPATH_SHOW = {
    'paths': MPath,
    'maps': MMap,
}
MP_SEP = ','

log = logging.getLogger('probert.multipath')


def _extract_mpath_data(cmd, show_verb):
    try:
        result = subprocess.run(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.DEVNULL)
    except (subprocess.CalledProcessError, FileNotFoundError):
        log.error('Failed to run cmd: %s', cmd)
        return []

    mptype = MPATH_SHOW[show_verb]
    data = result.stdout.decode('utf-8')
    result = []
    for line in data.splitlines():
        mp_dict = None
        try:
            field_vals = line.split(MP_SEP)
            log.debug('Extracted multipath %s fields: %s',
                      show_verb, field_vals)
            mp_dict = mptype(*field_vals)._asdict()
        except TypeError as e:
            log.debug(
                'Failed to parse multipath %s entry: %s: %s' % (show_verb,
                                                                line, e))
        if mp_dict:
            result.append(mp_dict)

    return result


def multipath_show_paths():
    path_format = MP_SEP.join(["%d", "%z", "%m", "%N", "%n", "%R", "%r", "%a"])
    cmd = ['multipathd', 'show', 'paths', 'raw', 'format', path_format]
    return _extract_mpath_data(cmd, 'paths')


def multipath_show_maps():
    maps_format = MP_SEP.join(["%w", "%d", "%N"])
    cmd = ['multipathd', 'show', 'maps', 'raw', 'format', maps_format]
    return _extract_mpath_data(cmd, 'maps')


def probe(context=None):
    """Query the multipath daemon for multipath maps and paths.

       This data is useful for determining whether a specific block
       device is part of a multipath and if so which device-mapper (dm)
       blockdevice should be used.

       This probe requires multipath module to be loaded and the multipath
       daemon to be running.
    """
    results = {}
    maps = multipath_show_maps()
    if maps:
        results.update({'maps': maps})
    paths = multipath_show_paths()
    if paths:
        results.update({'paths': paths})

    return results



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import subprocess

import pyudev

from probert.utils import (
    read_sys_block_size_bytes,
    sane_block_devices,
    )


log = logging.getLogger('probert.raid')

SUPPORTED_RAID_TYPES = ['raid0', 'raid1', 'raid5', 'raid6', 'raid10']


def mdadm_assemble(scan=True, ignore_errors=True):
    cmd = ['mdadm', '--detail', '--scan', '-v']
    try:
        subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError as e:
        log.error('Failed mdadm_assemble command %s: %s', cmd, e)
    except FileNotFoundError as e:
        log.error('Failed mdadm_assemble, mdadm command not found: %s', e)

    return


def get_mdadm_array_members(md_device):
    ''' extract array devices and spares from mdadm --detail --export output

    MD_LEVEL=raid5
    MD_DEVICES=3
    MD_METADATA=1.2
    MD_UUID=7fe1895e:34dcb6dc:d1bcbb9c:f3e05134
    MD_NAME=s1lp6:raid5-2406-2407-2408-2409
    MD_DEVICE_ev_dm_5_ROLE=spare
    MD_DEVICE_ev_dm_5_DEV=/dev/dm-5
    MD_DEVICE_ev_dm_3_ROLE=1
    MD_DEVICE_ev_dm_3_DEV=/dev/dm-3
    MD_DEVICE_ev_dm_4_ROLE=2
    MD_DEVICE_ev_dm_4_DEV=/dev/dm-4
    MD_DEVICE_ev_dm_2_ROLE=0
    MD_DEVICE_ev_dm_2_DEV=/dev/dm-2

    returns (['/dev/dm2', '/dev/dm-3', '/dev/dm-4'], ['/dev/dm-5'])
    '''
    cmd = ['mdadm', '--detail', '--export', md_device]
    try:
        result = subprocess.run(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.DEVNULL)
        output = result.stdout.decode('utf-8')
    except subprocess.CalledProcessError as e:
        log.error('failed to get detail for %s: %s', md_device, e)
        return ([], [])

    devices = {}
    roles = {}

    for line in output.splitlines():
        line = line.strip()
        if '=' not in line:
            continue
        k, v = line.split('=', 1)
        if k.startswith('MD_DEVICE_'):
            if k.endswith("_DEV"):
                dev_key = k[len('MD_DEVICE_'):-len('_DEV')]
                devices[dev_key] = v
            elif k.endswith("_ROLE"):
                dev_key = k[len('MD_DEVICE_'):-len('_ROLE')]
                roles[dev_key] = v

    actives = []
    spares = []

    for dev_key, devname in devices.items():
        if roles.get(dev_key) == 'spare':
            spares.append(devname)
        else:
            actives.append(devname)
    return (sorted(actives), sorted(spares))


def probe(context=None, report=False):
    """Initiate an mdadm assemble to awaken existing MDADM devices.
       For each md block device, extract required information needed
       to describe the array for recreation or reuse as needed.

       mdadm tooling provides information about the raid type,
       the members, the size, the name, uuids, metadata version.
    """
    mdadm_assemble()

    # ignore passed context, must read udev after assembling mdadm devices
    context = pyudev.Context()

    raids = {}
    for device in sane_block_devices(context):
        if device.get('DEVTYPE') != 'disk':
            continue
        devname = device['DEVNAME']
        if not os.path.basename(devname).startswith('md'):
            continue
        if 'MD_CONTAINER' in device:
            cfg = dict(device)
            cfg.update({
                'raidlevel': device['MD_LEVEL'],
                'container': device['MD_CONTAINER'],
                'size': str(read_sys_block_size_bytes(devname)),
                })
            raids[devname] = cfg
        else:
            devices, spares = get_mdadm_array_members(devname)
            cfg = dict(device)
            if device.get('MD_METADATA') == 'imsm':
                # All disks in a imsm container show up as spares, in some
                # sense because they are not "used" by the container (there is
                # a concept of a spare drive in a container -- where there is a
                # drive in the container that is not part of a volume/subarray
                # within it -- but this is a fairly ephemeral concept which
                # doesn't survive a reboot, so we don't account for that
                # here). We don't care about that though and just record all
                # component disks as active.
                devices = devices + spares
                spares = []
            cfg.update({
                'raidlevel': device['MD_LEVEL'],
                'devices': devices,
                'spare_devices': spares,
                'size': str(read_sys_block_size_bytes(devname)),
                })
            raids[devname] = cfg

    return raids



# Copyright 2020 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import platform
import pyudev
import re
import subprocess

from probert.utils import sane_block_devices


log = logging.getLogger('probert.dasd')


def _dasd_view_dec_pattern(label):
    return r"^{}\s+:\shex\s\w+\s+dec\s(?P<value>\d+)$".format(
        re.escape(label))


DASD_FORMAT = r"^format\s+:.+\s+(?P<value>\w+\s\w+)$"
DASD_BLKSIZE = _dasd_view_dec_pattern("blocksize")
DASD_CYLINDERS = _dasd_view_dec_pattern("number of cylinders")
DASD_TRACKS_PER_CYLINDER = _dasd_view_dec_pattern("tracks per cylinder")
DASD_TYPE = r"^type\s+:\s(?P<value>[A-Za-z]+)\s*$"


def find_val(regex, content):
    m = re.search(regex, content, re.MULTILINE)
    if m is not None:
        return m.group("value")


def find_val_int(regex, content):
    v = find_val(regex, content)
    if v is not None:
        return int(v)


def disk_format(dasdview_output):
    """ Read and return specified device "disk_layout" value.

    :returns: string: One of ['cdl', 'ldl', 'not-formatted'].
    :raises: ValueError if dasdview result missing 'format' section.

    """
    if not dasdview_output:
        return

    mapping = {
       'cdl formatted': 'cdl',
       'ldl formatted': 'ldl',
       'not formatted': 'not-formatted',
    }
    diskfmt = find_val(DASD_FORMAT, dasdview_output)
    if diskfmt is not None:
        return mapping.get(diskfmt.lower())


def dasdview(devname):
    ''' Run dasdview on devname and return the output.

    dasdview --extended has 3 sections
    general (2:6), geometry (8:12), extended (14:)

    '''
    if not os.path.exists(devname):
        raise ValueError("Invalid dasd device name: '%s'" % devname)

    cmd = ['dasdview', '--extended', devname]
    try:
        result = subprocess.run(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.DEVNULL)
    except (subprocess.CalledProcessError, FileNotFoundError):
        log.error('Failed to run cmd: %s', cmd)
        return None

    return result.stdout.decode('utf-8')


def get_dasd_info(device):
    """ from a udev blockdev device entry, return all required dasd info

    """
    name = device.get('DEVNAME')
    device_id = device.get('ID_PATH', '').replace('ccw-', '')

    dasdview_output = dasdview(name)
    diskfmt = disk_format(dasdview_output)
    blksize = find_val_int(DASD_BLKSIZE, dasdview_output)
    type = find_val(DASD_TYPE, dasdview_output)

    cylinders = find_val_int(DASD_CYLINDERS, dasdview_output)
    tracks_per_cylinder = find_val_int(
        DASD_TRACKS_PER_CYLINDER, dasdview_output)

    if not all([name, device_id, diskfmt, blksize]):
        vals = ("name=%s device_id=%s format=%s blksize=%s" % (
                name, device_id, diskfmt, blksize))
        log.debug('Failed to probe some DASD values: %s', vals)
        return None

    return {
        'blocksize': blksize,
        'cylinders': cylinders,
        'device_id': device_id,
        'disk_layout': diskfmt,
        'name': name,
        'tracks_per_cylinder': tracks_per_cylinder,
        'type': type,
        }


def probe(context=None):
    """Examine all dasd devices present and extract configuration attributes

       This data is useful for determining if the dasd device has been
       formatted, if so what the block size, the partition layout used
       and the s390x device_id used to uniquely identify the device.
    """
    log.debug('Probing DASD devies')
    machine = platform.machine()
    if machine != "s390x":
        log.debug('DASD devices only present on s390x, arch=%s', machine)
        return {}

    dasds = {}
    if not context:
        context = pyudev.Context()

    virtio_major = None
    for line in open('/proc/devices'):
        if line.endswith('virtblk\n'):
            virtio_major = line.split()[0]

    log.debug("found MAJOR for virtblk: %s", virtio_major)

    for device in sane_block_devices(context):
        # ignore partitions
        if 'PARTN' in device:
            continue
        # dasd devices have MAJOR 94
        if device['MAJOR'] == "94":
            try:
                dasd_info = get_dasd_info(device)
            except ValueError as e:
                log.error(
                    'Error probing dasd device %s: %s', device['DEVNAME'], e)
                dasd_info = None

            if dasd_info:
                dasds[device['DEVNAME']] = dasd_info
        elif device['MAJOR'] == virtio_major:
            # a dasd can be passed to a VM via virtio, in which case
            # there is no device id and dasdview/dasdmft do not work
            # but it must still be formatted with vtoc, so we report
            # it here. The only way I can find to detect such a device
            # is that "fdasd -i" on the device is successful.
            result = subprocess.run(
                ['fdasd', '-i', device['DEVNAME']],
                stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            log.debug(
                "fasd -i %s returned %s",
                device['DEVNAME'], result.returncode)
            if result.returncode == 0:
                dasds[device['DEVNAME']] = {
                    'name': device['DEVNAME'],
                    'type': 'virt',
                    }

    return dasds



import os

TOP_DIR = os.path.join('/'.join(__file__.split('/')[:-3]))
TEST_DATA = os.path.join(TOP_DIR, 'probert', 'tests', 'data')
FAKE_PROBE_ALL_JSON = os.path.join(TEST_DATA, 'fake_probe_all.json')



from copy import deepcopy
import glob
import itertools
import os
import re

import pyudev

NET_CONFIG_OPTIONS = [
    "address", "netmask", "broadcast", "network", "metric", "gateway",
    "pointtopoint", "media", "mtu", "hostname", "leasehours", "leasetime",
    "vendor", "client", "bootfile", "server", "hwaddr", "provider", "frame",
    "netnum", "endpoint", "local", "ttl",
]

NET_CONFIG_COMMANDS = [
    "pre-up", "up", "post-up", "down", "pre-down", "post-down",
]

NET_CONFIG_BRIDGE_OPTIONS = [
    "bridge_ageing", "bridge_bridgeprio", "bridge_fd", "bridge_gcinit",
    "bridge_hello", "bridge_maxage", "bridge_maxwait", "bridge_stp",
]

# sysfs size attribute is always in 512-byte units
# https://github.com/torvalds/linux/blob/6f0d349d922ba44e4348a17a78ea51b7135965b1/include/linux/types.h#L125
SECTOR_SIZE_BYTES = 512


# from juju-deployer utils.relation_merge
def dict_merge(onto, source):
    target = deepcopy(onto)
    # Support list of relations targets
    if isinstance(onto, list) and isinstance(source, list):
        target.extend(source)
        return target
    for (key, value) in source.items():
        if key in target:
            if isinstance(target[key], dict) and isinstance(value, dict):
                target[key] = dict_merge(target[key], value)
            elif isinstance(target[key], list) and isinstance(value, list):
                target[key] = list(set(target[key] + value))
        else:
            target[key] = value
    return target


if pyudev.__version_info__ < (0, 18):
    def udev_get_attributes(device):
        r = {}
        for key in device.attributes:
            val = device.attributes.get(key)
            if isinstance(val, bytes):
                val = val.decode('utf-8', 'replace')
            r[key] = val
        return r
else:
    def udev_get_attributes(device):
        r = {}
        for key in device.attributes.available_attributes:
            val = device.attributes.get(key)
            if isinstance(val, bytes):
                val = val.decode('utf-8', 'replace')
            r[key] = val
        return r


# split lists into N lists by predicate
def partitionn2(items, predicate=int, n=2):
    return ((lambda i, tee: (item for pred, item in tee if pred == i))(x, t)
            for x, t in enumerate(itertools.tee(((predicate(item), item)
                                  for item in items), n)))


# unpack generators into key, value pair
# where key is first item (list[0]) and
# value is remainder (list[1:])
def partition_to_pair(input):
    """Unpack a partition into a tuple of (first partition, second partition)

    param: partition iterator from partitionn2
    """
    items = input.split()
    partitions = partitionn2(items=items,
                             predicate=lambda x: items.index(x) != 0,
                             n=2)
    data = [list(p) for p in partitions]
    [key], value = data
    return key, value


def disentagle_data_from_whitespace(data):
    # disentagle the data from whitespace
    return [x.split(';')[0].strip() for x in data.split('\n')
            if len(x)]


def dictify_lease(lease):
    """Transform lease string into dictionary of attributes

    params: lease: string if a dhcp lease structure { to }
    """
    lease_dict = {}
    options = {}
    for line in disentagle_data_from_whitespace(lease):
        if len(line) <= 0:
            continue

        key, value = partition_to_pair(line)
        if key == 'option':
            options.update({value[0]: value[1]})
        else:
            value = " ".join(value)
            lease_dict.update({key: value})

    lease_dict.update({'options': options})
    return lease_dict


def parse_dhclient_leases_file(leasedata):
    """Parses dhclient leases file data, returning dictionary of leases

    :param leasesdata: string of lease data read from leases file
    """
    return [dictify_lease(lease) for lease in
            re.findall(r'{([^{}]*)}', leasedata.replace('"', ''))]


def parse_networkd_lease_file(leasedata):
    """Parses systemd/networkd/netif lease data, returns dict"""
    lease = {}
    for line in leasedata.split('\n'):
        if line.startswith('#') or len(line) < 1:
            continue
        keyvalue = line.split('=')
        lease[keyvalue[0].lower()] = keyvalue[1]
    return lease


def get_dhclient_d():
    # find lease files directory
    supported_dirs = ["/var/lib/dhcp", "/var/lib/dhclient"]
    for d in supported_dirs:
        if os.path.exists(d):
            return d
    return None


def parse_etc_network_interfaces(ifaces, contents, path):
    """Parses the file contents, placing result into ifaces.

    :param ifaces: interface dictionary
    :param contents: contents of interfaces file
    :param path: directory interfaces file was located
    """
    currif = None
    src_dir = path
    for line in contents.splitlines():
        line = line.strip()
        if line.startswith('#'):
            continue
        split = line.split(' ')
        option = split[0]
        if option == "source-directory":
            src_dir = os.path.join(path, split[1])
        elif option == "source":
            src_path = os.path.join(src_dir, split[1])
            for src_file in glob.glob(src_path):
                with open(src_file, "r") as fp:
                    src_data = fp.read().strip()
                parse_etc_network_interfaces(
                    ifaces, src_data,
                    os.path.dirname(os.path.abspath(src_file)))
        elif option == "auto":
            for iface in split[1:]:
                if iface not in ifaces:
                    ifaces[iface] = {}
                ifaces[iface]['auto'] = True
        elif option == "iface":
            iface, family, method = split[1:4]
            if iface not in ifaces:
                ifaces[iface] = {}
            elif 'family' in ifaces[iface]:
                raise Exception("Cannot define %s interface again.")
            ifaces[iface]['family'] = family
            ifaces[iface]['method'] = method
            currif = iface
        elif option == "hwaddress":
            ifaces[currif]['hwaddress'] = split[1]
        elif option in NET_CONFIG_OPTIONS:
            ifaces[currif][option] = split[1]
        elif option in NET_CONFIG_COMMANDS:
            if option not in ifaces[currif]:
                ifaces[currif][option] = []
            ifaces[currif][option].append(' '.join(split[1:]))
        elif option.startswith('dns-'):
            if 'dns' not in ifaces[currif]:
                ifaces[currif]['dns'] = {}
            if option == 'dns-search':
                ifaces[currif]['dns']['search'] = []
                for domain in split[1:]:
                    ifaces[currif]['dns']['search'].append(domain)
            elif option == 'dns-nameservers':
                ifaces[currif]['dns']['nameservers'] = []
                for server in split[1:]:
                    ifaces[currif]['dns']['nameservers'].append(server)
        elif option.startswith('bridge_'):
            if 'bridge' not in ifaces[currif]:
                ifaces[currif]['bridge'] = {}
            if option in NET_CONFIG_BRIDGE_OPTIONS:
                bridge_option = option.replace('bridge_', '')
                ifaces[currif]['bridge'][bridge_option] = split[1]
            elif option == "bridge_ports":
                ifaces[currif]['bridge']['ports'] = []
                for iface in split[1:]:
                    ifaces[currif]['bridge']['ports'].append(iface)
            elif option == "bridge_hw" and split[1].lower() == "mac":
                ifaces[currif]['bridge']['mac'] = split[2]
            elif option == "bridge_pathcost":
                if 'pathcost' not in ifaces[currif]['bridge']:
                    ifaces[currif]['bridge']['pathcost'] = {}
                ifaces[currif]['bridge']['pathcost'][split[1]] = split[2]
            elif option == "bridge_portprio":
                if 'portprio' not in ifaces[currif]['bridge']:
                    ifaces[currif]['bridge']['portprio'] = {}
                ifaces[currif]['bridge']['portprio'][split[1]] = split[2]
    for iface in ifaces.keys():
        if 'auto' not in ifaces[iface]:
            ifaces[iface]['auto'] = False


def read_sys_block_size_bytes(device):
    """ /sys/class/block/<device>/size and return integer value in bytes"""
    device_dir = os.path.join('/sys/class/block', os.path.basename(device))
    blockdev_size = os.path.join(device_dir, 'size')
    with open(blockdev_size) as d:
        size = int(d.read().strip()) * SECTOR_SIZE_BYTES

    return size


def read_sys_block_slaves(device):
    device_dir = os.path.join('/sys/class/block', os.path.basename(device))
    return os.listdir(os.path.join(device_dir, 'slaves'))


def sane_block_devices(context):
    for device in context.list_devices(subsystem='block'):
        if "MAJOR" not in device:
            # Shouldn't happen but apparently does! (LP: #1868109)
            continue
        yield device



import mock
import subprocess
import testtools

from probert import multipath
from probert.tests.helpers import random_string

MP_SEP = multipath.MP_SEP


class TestMultipath(testtools.TestCase):

    @mock.patch('probert.multipath.subprocess.run')
    def test_multipath_show_paths(self, m_run):
        mp_out = MP_SEP.join([random_string() for x in range(0, 8)])
        cp = subprocess.CompletedProcess(args=['foo'], returncode=0,
                                         stdout=mp_out.encode('utf-8'),
                                         stderr="")
        m_run.return_value = cp
        expected_result = [multipath.MPath(*mp_out.split(MP_SEP))._asdict()]
        result = multipath.multipath_show_paths()
        self.assertEqual(expected_result, result)

    @mock.patch('probert.multipath.subprocess.run')
    def test_multipath_show_paths_serial_with_spaces(self, m_run):
        mp_out = MP_SEP.join(['sda', 'IPR-0 1234567890'] +
                             [random_string() for x in range(0, 6)])
        cp = subprocess.CompletedProcess(args=['foo'], returncode=0,
                                         stdout=mp_out.encode('utf-8'),
                                         stderr="")
        m_run.return_value = cp
        expected_result = [multipath.MPath(*mp_out.split(MP_SEP))._asdict()]
        result = multipath.multipath_show_paths()
        self.assertEqual(expected_result, result)

    @mock.patch('probert.multipath.subprocess.run')
    def test_multipath_show_paths_skips_unparsable_output(self, m_run):
        lines = [
            MP_SEP.join([random_string() for x in range(0, 8)]),
            "",
            MP_SEP.join([random_string() for x in range(0, 8)]),
        ]
        mp_out = "\n".join(lines)
        cp = subprocess.CompletedProcess(args=['foo'], returncode=0,
                                         stdout=mp_out.encode('utf-8'),
                                         stderr="")
        m_run.return_value = cp
        expected_result = [multipath.MPath(*lines[0].split(MP_SEP))._asdict(),
                           multipath.MPath(*lines[2].split(MP_SEP))._asdict()]
        result = multipath.multipath_show_paths()
        self.assertEqual(expected_result, result)

    @mock.patch('probert.multipath.subprocess.run')
    def test_multipath_show_maps(self, m_run):
        mp_out = MP_SEP.join([random_string() for x in range(0, 3)])
        cp = subprocess.CompletedProcess(args=['foo'], returncode=0,
                                         stdout=mp_out.encode('utf-8'),
                                         stderr="")
        m_run.return_value = cp
        expected_result = [multipath.MMap(*mp_out.split(MP_SEP))._asdict()]
        result = multipath.multipath_show_maps()
        self.assertEqual(expected_result, result)

    @mock.patch('probert.multipath.subprocess.run')
    def test_multipath_show_maps_skips_unparsable_output(self, m_run):
        lines = [
            MP_SEP.join([random_string() for x in range(0, 3)]),
            "",
            MP_SEP.join([random_string() for x in range(0, 3)]),
        ]
        mp_out = "\n".join(lines)
        cp = subprocess.CompletedProcess(args=['foo'], returncode=0,
                                         stdout=mp_out.encode('utf-8'),
                                         stderr="")
        m_run.return_value = cp
        expected_result = [multipath.MMap(*lines[0].split(MP_SEP))._asdict(),
                           multipath.MMap(*lines[2].split(MP_SEP))._asdict()]
        result = multipath.multipath_show_maps()
        self.assertEqual(expected_result, result)

    @mock.patch('probert.multipath.subprocess.run')
    def test_multipath_extract_returns_empty_list_on_err(self, m_run):
        m_run.side_effect = subprocess.CalledProcessError(cmd=["my cmd"],
                                                          returncode=1)
        result = multipath.multipath_show_paths()
        self.assertEqual([], result)

    @mock.patch('probert.multipath.multipath_show_paths')
    @mock.patch('probert.multipath.multipath_show_maps')
    def test_multipath_probe_collects_maps_and_paths(self, m_maps, m_paths):
        path_string = MP_SEP.join([random_string() for x in range(0, 8)])
        paths = multipath.MPath(*path_string.split(MP_SEP))._asdict()
        maps_string = MP_SEP.join([random_string() for x in range(0, 3)])
        maps = multipath.MMap(*maps_string.split(MP_SEP))._asdict()
        m_maps.return_value = [maps]
        m_paths.return_value = [paths]
        result = multipath.probe()
        self.assertDictEqual({'maps': [maps], 'paths': [paths]}, result)



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import subprocess

import pyudev

from probert.utils import sane_block_devices


log = logging.getLogger('probert.dmcrypt')


def dmsetup_info(devname):
    ''' returns dict of info about device mapper dev.

    {'blkdevname': 'dm-0',
     'blkdevs_used': 'sda5',
     'name': 'sda5_crypt',
     'subsystem': 'CRYPT',
     'uuid': 'CRYPT-LUKS1-2b370697149743b0b2407d11f88311f1-sda5_crypt'
    }
    '''
    _SEP = '='
    fields = ('name,uuid,blkdevname,blkdevs_used,subsystem'.split(','))
    try:
        output = subprocess.check_output(
            ['sudo', 'dmsetup', 'info', devname, '-C', '-o',
             ','.join(fields), '--noheading', '--separator', _SEP])
    except subprocess.CalledProcessError as e:
        log.error('Failed to probe dmsetup info:', e)
        return None
    values = output.decode('utf-8').strip().split(_SEP)
    info = dict(zip(fields, values))
    return info


def probe(context=None, report=False):
    """ Probing for dm_crypt devices requires running dmsetup info commands
        to collect how a particular dm-X device is composed.
    """
    # ignore supplied context, we need to read udev after scan/vgchange
    context = pyudev.Context()

    crypt_devices = {}

    # look for block devices with DM_UUID and CRYPT; these are crypt devices
    for device in sane_block_devices(context):
        if 'DM_UUID' in device and device['DM_UUID'].startswith('CRYPT'):
            devname = device['DEVNAME']
            dm_info = dmsetup_info(devname)
            crypt_devices[dm_info['name']] = dm_info

    return crypt_devices



# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from logging.handlers import TimedRotatingFileHandler


def setup_logger(name=__name__):
    LOGDIR = "logs"
    LOGFILE = os.path.join(LOGDIR, "debug.log")
    if not os.path.isdir(LOGDIR):
        os.makedirs(LOGDIR)
    log = TimedRotatingFileHandler(LOGFILE,
                                   when='D',
                                   interval=1,
                                   backupCount=7)
    log.setLevel('DEBUG')
    log.setFormatter(logging.Formatter(
        "%(asctime)s "
        "%(name)s:%(lineno)d %(message)s",
        datefmt='%m/%d %H:%M'))
    log_filter = logging.Filter(name='probert')
    log.addFilter(log_filter)

    logger = logging.getLogger('')
    logger.setLevel('DEBUG')
    logger.addHandler(log)
    return logger



import json
import mock
import subprocess
import testtools
import textwrap

from probert import lvm
from probert.tests.helpers import random_string

CONTEXT = [
  {
    "DEVNAME": "/dev/dm-0",
    "DEVTYPE": "disk",
    "DM_LV_NAME": "lv1",
    "DM_NAME": "vg1-lv1",
    "DM_UUID": "LVM-IBOfU1ELB4dehjpX2wy3BtkD504ARo0oclQZnCRnv2TGeopW5eZiP",
    "DM_VG_NAME": "vg1",
    "attrs": {
      "size": "1073741824",
    }
  }, {
    "DEVNAME": "/dev/vda5",
    "DEVTYPE": "partition",
    "ID_FS_TYPE": "LVM2_member",
    "USEC_INITIALIZED": "130182115",
    "attrs": {
      "size": "2147483648",
    }
  }, {
    "DEVNAME": "/dev/vda6",
    "DEVTYPE": "partition",
    "ID_FS_TYPE": "LVM2_member",
    "attrs": {
      "size": "3221225472",
    }
  },
]

VGS_REPORT = [
    {
        "vg_name": "vg1",
        "pv_name": "/dev/vda5",
        "pv_uuid": "46GvlZ-5UAL-tvnf-AX60-3pwg-A9p3-5pBAq6",
        "vg_size": "206145847296B"
    }, {
        "vg_name": "vg1",
        "pv_name": "/dev/vda6",
        "pv_uuid": "J3bQo5-scWX-fIwg-Je9J-LkRC-rkLp-1X6KPg",
        "vg_size": "206145847296B"
    },
]

CONTEXT_DUPES = 2 * CONTEXT + [
  {
    "DEVNAME": "/dev/dm-1",
    "DEVTYPE": "disk",
    "DM_LV_NAME": "lv2",
    "DM_NAME": "vg1-lv2",
    "DM_UUID": "LVM-IBOfU1ELB4dehjpX2wy3BtkD504ARo0oclQZnCRnv2TGeopW5eZiP",
    "DM_VG_NAME": "vg1",
    "attrs": {
      "size": "1073741824",
    }
  },
]
VGS_REPORT_DUPES = 2 * VGS_REPORT


@mock.patch('probert.lvm.subprocess.run')
class TestLvm(testtools.TestCase):

    def test__lvm_report_returns_empty_list_on_err(self, m_run):
        m_run.side_effect = subprocess.CalledProcessError(
            cmd=[random_string()], returncode=1)
        result = lvm._lvm_report(random_string(), random_string())
        self.assertEqual([], result)

    def test__lvm_report_returns_empty_list_on_no_output(self, m_run):
        cmd_out = ""
        cp = subprocess.CompletedProcess(args=[random_string()], returncode=0,
                                         stdout=cmd_out.encode('utf-8'),
                                         stderr="")
        m_run.return_value = cp
        self.assertEqual([], lvm._lvm_report(random_string(), random_string()))

    def test__lvm_report_returns_empty_list_on_invalid_json(self, m_run):
        cmd_out = "This is not json"
        cp = subprocess.CompletedProcess(args=[random_string()], returncode=0,
                                         stdout=cmd_out.encode('utf-8'),
                                         stderr="")
        m_run.return_value = cp
        self.assertEqual([], lvm._lvm_report(random_string(), random_string()))

    def test__lvm_report_returns_found_reports(self, m_run):
        report_key = random_string()
        report_data = [{random_string(): random_string()}]
        cmd_out = json.dumps({"report": [{report_key: report_data}]})
        cp = subprocess.CompletedProcess(args=[random_string()], returncode=0,
                                         stdout=cmd_out.encode('utf-8'),
                                         stderr="")
        m_run.return_value = cp
        expected_result = report_data
        result = lvm._lvm_report(random_string(), report_key)
        self.assertEqual(expected_result, result)

    def test__lvm_report_returns_specific_reports(self, m_run):
        report_key = 'report-1'
        report_data = [{random_string(): random_string()}]
        extra_data1 = [list(range(0, 10))]
        extra_data2 = ""
        cmd_out = json.dumps({
            "report": [
                {report_key: report_data},
                {random_string(): extra_data1},
                {random_string(): extra_data2},
             ]
        })
        cp = subprocess.CompletedProcess(args=[random_string()], returncode=0,
                                         stdout=cmd_out.encode('utf-8'),
                                         stderr="")
        m_run.return_value = cp
        expected_result = report_data
        result = lvm._lvm_report(random_string(), report_key)
        self.assertEqual(expected_result, result)

    @mock.patch('probert.lvm._lvm_report')
    def test_probe_pvs_report_calls_pvs(self, m_lvreport, m_run):
        lvm.probe_pvs_report()
        m_lvreport.assert_called_with(['pvs', '--reportformat=json'], 'pv')

    @mock.patch('probert.lvm._lvm_report')
    def test_probe_vgs_report_calls_vgs(self, m_lvreport, m_run):
        lvm.probe_vgs_report()
        m_lvreport.assert_called_with(
            ['vgs', '--reportformat=json', '--units=B',
             '-o', 'vg_name,pv_name,pv_uuid,vg_size'], 'vg')

    @mock.patch('probert.lvm._lvm_report')
    def test_probe_lvs_report_calls_lvs(self, m_lvreport, m_run):
        lvm.probe_lvs_report()
        m_lvreport.assert_called_with(['lvs'], 'lv')

    @mock.patch('probert.lvm.os.environ')
    def test_lvmetad_running(self, m_env, m_run):
        m_env.return_value = {'LVM_LVMETA_D_PIDFILE': '/run/lvmetad.pid'}
        self.assertTrue(lvm.lvmetad_running())

    @mock.patch('probert.lvm.os.environ.get')
    @mock.patch('probert.lvm.os.path.exists')
    def test_lvmetad_running_env_get(self, m_path, m_env, m_run):
        pidfile = random_string()
        m_env.return_value = pidfile
        m_path.return_value = True
        self.assertTrue(lvm.lvmetad_running())
        m_env.assert_called_with('LVM_LVMETAD_PIDFILE', '/run/lvmetad.pid')
        m_path.assert_called_with(pidfile)

    @mock.patch('probert.lvm.os.environ')
    @mock.patch('probert.lvm.os.path.exists')
    def test_lvmetad_running_env_empty(self, m_path, m_env, m_run):
        pidfile = '/run/lvmetad.pid'
        m_env.get.return_value = pidfile
        m_path.return_value = True
        self.assertTrue(lvm.lvmetad_running())
        m_path.assert_called_with(pidfile)

    @mock.patch('probert.lvm.os.environ')
    @mock.patch('probert.lvm.os.path.exists')
    def test_lvmetad_running_false_if_path_fails(self, m_path, m_env, m_run):
        m_path.return_value = False
        self.assertFalse(lvm.lvmetad_running())

    @mock.patch('probert.lvm.lvmetad_running')
    def test_lvm_scan(self, m_metad, m_run):
        m_metad.return_value = True
        lvm.lvm_scan()
        self.assertEqual(2, m_metad.call_count)
        m_run.assert_has_calls([
          mock.call(['pvscan', '--cache'],
                    stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL),
          mock.call(['vgscan', '--cache'],
                    stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)])

    @mock.patch('probert.lvm.lvmetad_running')
    def test_lvm_scan_handles_errors(self, m_metad, m_run):
        m_metad.return_value = True
        m_run.side_effect = iter([
            subprocess.CalledProcessError(cmd="foo", returncode=1),
            subprocess.CalledProcessError(cmd="foo", returncode=1),
        ])
        lvm.lvm_scan()
        self.assertEqual(2, m_metad.call_count)
        m_run.assert_has_calls([
          mock.call(['pvscan', '--cache'],
                    stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL),
          mock.call(['vgscan', '--cache'],
                    stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)])

    def test_activate_volgroups(self, m_run):
        lvm.activate_volgroups()
        m_run.assert_has_calls([
          mock.call(['vgchange', '--activate=y'], check=False,
                    stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)])

    def test_extract_lvm_volgroup(self, m_run):
        input_data = json.loads(
            textwrap.dedent("""
                [
                    {"vg_name": "vg0",
                     "pv_name": "/dev/md0",
                     "pv_uuid": "p3oDow-dRHp-L8jq-t6gQ-67tv-B8B6-JWLKZP",
                     "vg_size": "1000B"},
                    {"vg_name": "vg0",
                     "pv_name": "/dev/md1",
                     "pv_uuid": "pRR5Zn-c4a9-teVZ-TFaU-yDxf-FSDo-cORcEq",
                     "vg_size": "21449670656B"},
                    {"vg_name": "vg0",
                     "pv_name": "/dev/md2",
                     "pv_uuid": "Xsjd5a-c4a9-teVZ-TFaU-yDxf-FSDo-cORcEq",
                     "vg_size": "doesnt_end_with_B_"}
                ]"""))
        self.assertEqual(
            ("vg0", {'name': "vg0",
                     'devices': sorted(['/dev/md0', '/dev/md1', '/dev/md2']),
                     'size': '21449670656B'}),
            lvm.extract_lvm_volgroup('vg0', input_data))

    def test_extract_lvm_volgroup_no_size_set_to_zero_bytes(self, m_run):
        input_data = json.loads(
            textwrap.dedent("""
                [
                    {"vg_name": "vg0",
                     "pv_name": "/dev/md0",
                     "pv_uuid": "p3oDow-dRHp-L8jq-t6gQ-67tv-B8B6-JWLKZP",
                     "vg_size": null},
                    {"vg_name": "vg0",
                     "pv_name": "/dev/md1",
                     "pv_uuid": "pRR5Zn-c4a9-teVZ-TFaU-yDxf-FSDo-cORcEq",
                     "vg_size": null}
                ]"""))
        self.assertEqual(
            ("vg0", {'name': "vg0",
                     'devices': sorted(['/dev/md0', '/dev/md1']),
                     'size': '0B'}),
            lvm.extract_lvm_volgroup('vg0', input_data))

    @mock.patch('probert.lvm.read_sys_block_size_bytes')
    def test_extract_lvm_partition(self, m_size, m_run):
        size = 100000000
        m_size.return_value = size
        input_data = {
            'DEVNAME': '/dev/dm-2',
            'DM_NAME': 'ubuntu--vg-my--storage',
            'DM_LV_NAME': 'my-storage',
            'DM_VG_NAME': 'ubuntu-vg',
        }
        self.assertEqual(
            ('ubuntu-vg/my-storage',
             {'fullname': 'ubuntu-vg/my-storage', 'name': 'my-storage',
              'volgroup': 'ubuntu-vg', 'size': "%sB" % size}),
            lvm.extract_lvm_partition(input_data))
        m_size.assert_called_with('/dev/dm-2')

    @mock.patch('probert.lvm.read_sys_block_size_bytes')
    @mock.patch('probert.lvm.activate_volgroups')
    @mock.patch('probert.lvm.lvm_scan')
    @mock.patch('probert.lvm.sane_block_devices')
    @mock.patch('probert.lvm.probe_vgs_report')
    def test_probe(self, m_vgs, m_blockdevs, m_scan, m_activate, m_size,
                   m_run):
        size = 1000
        m_size.return_value = size
        m_blockdevs.return_value = CONTEXT
        m_vgs.return_value = VGS_REPORT

        expected_result = {
            'logical_volumes': {
                'vg1/lv1': {
                    'fullname': 'vg1/lv1',
                    'name': 'lv1',
                    'size': '1000B',
                    'volgroup': 'vg1'
                }
            },
            'physical_volumes': {
                'vg1': ['/dev/vda5', '/dev/vda6']
            },
            'volume_groups': {
                'vg1': {
                    'devices': ['/dev/vda5', '/dev/vda6'],
                    'name': 'vg1',
                    'size': '206145847296B'
                }
            }
        }
        self.assertEqual(expected_result, lvm.probe())

    @mock.patch('probert.lvm.read_sys_block_size_bytes')
    @mock.patch('probert.lvm.activate_volgroups')
    @mock.patch('probert.lvm.lvm_scan')
    @mock.patch('probert.lvm.sane_block_devices')
    @mock.patch('probert.lvm.probe_vgs_report')
    def test_probe_skip_dupes(self, m_vgs, m_blockdevs, m_scan, m_activate,
                              m_size, m_run):
        size = 1000
        m_size.return_value = size
        m_blockdevs.return_value = CONTEXT_DUPES
        m_vgs.return_value = VGS_REPORT_DUPES

        expected_result = {
            'logical_volumes': {
                'vg1/lv1': {
                    'fullname': 'vg1/lv1',
                    'name': 'lv1',
                    'size': '1000B',
                    'volgroup': 'vg1'
                },
                'vg1/lv2': {
                    'fullname': 'vg1/lv2',
                    'name': 'lv2',
                    'size': '1000B',
                    'volgroup': 'vg1'
                },
            },
            'physical_volumes': {
                'vg1': ['/dev/vda5', '/dev/vda6']
            },
            'volume_groups': {
                'vg1': {
                    'devices': ['/dev/vda5', '/dev/vda6'],
                    'name': 'vg1',
                    'size': '206145847296B'
                }
            }
        }
        self.assertEqual(expected_result, lvm.probe())


# vi: ts=4 expandtab syntax=python



import mock
import subprocess
import testtools

from probert import dasd
from probert.tests import fakes
from probert.tests.helpers import random_string


# The tests parse canned dasdview output, and to be able to write
# tests one needs to simply know what the correct parsed values
# are. These helper functions help make the this dependence on outside
# knowledge a little clearer:


def update_probe_data(base, **kw):
    r = base.copy()
    r.update(kw)
    return r


expected_probe_data = {
    '/dev/dasdd': {
        'blocksize': 4096,
        'cylinders': 30051,
        'disk_layout': 'cdl',
        'name': '/dev/dasdd',
        'tracks_per_cylinder': 15,
        'type': 'ECKD',
        },
    '/dev/dasde': {
        'blocksize': 512,
        'cylinders': 10017,
        'disk_layout': 'not-formatted',
        'name': '/dev/dasde',
        'tracks_per_cylinder': 15,
        'type': 'ECKD',
        },
    }


class TestDasd(testtools.TestCase):

    def _load_test_data(self, data_fname):
        testfile = fakes.TEST_DATA + '/' + data_fname
        with open(testfile, 'r') as fh:
            return fh.read()

    @mock.patch('probert.dasd.os.path.exists')
    @mock.patch('probert.dasd.subprocess.run')
    def test_dasdview_returns_stdout(self, m_run, m_exists):
        devname = random_string()
        dasdview_out = random_string()
        cp = subprocess.CompletedProcess(args=['foo'], returncode=0,
                                         stdout=dasdview_out.encode('utf-8'),
                                         stderr="")
        m_run.return_value = cp
        m_exists.return_value = True
        result = dasd.dasdview(devname)
        self.assertEqual(dasdview_out, result)
        m_run.assert_called_with(['dasdview', '--extended', devname],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.DEVNULL)

    @mock.patch('probert.dasd.os.path.exists')
    @mock.patch('probert.dasd.subprocess.run')
    def test_dasdview_raises_valueerror(self, m_run, m_exists):
        devname = random_string()
        m_exists.return_value = False
        self.assertRaises(ValueError, dasd.dasdview, devname)
        self.assertEqual(0, m_run.call_count)

    @mock.patch('probert.dasd.os.path.exists')
    @mock.patch('probert.dasd.subprocess.run')
    def test_dasdview_returns_none_on_subprocess_error(self, m_run, m_exists):
        devname = random_string()
        m_exists.return_value = True
        m_run.side_effect = subprocess.CalledProcessError(
            cmd=[random_string()], returncode=1)
        self.assertEqual(None, dasd.dasdview(devname))

    def test_dasd_parses_blocksize(self):
        self.assertEqual(
            4096,
            dasd.find_val_int(
                dasd.DASD_BLKSIZE, self._load_test_data('dasdd.view')))

    def test_dasd_blocksize_returns_none_on_invalid_output(self):
        self.assertIsNone(
            dasd.find_val_int(
                dasd.DASD_BLKSIZE, random_string()))

    def test_dasd_parses_disk_format(self):
        self.assertEqual('cdl',
                         dasd.disk_format(self._load_test_data('dasdd.view')))
        self.assertEqual('not-formatted',
                         dasd.disk_format(self._load_test_data('dasde.view')))

    def test_dasd_parses_disk_format_ldl(self):
        output = "format : hex 1 dec 1 LDL formatted"
        self.assertEqual('ldl', dasd.disk_format(output))

    def test_dasd_disk_format_returns_none_on_invalid_output(self):
        self.assertIsNone(dasd.disk_format(random_string()))

    @mock.patch('probert.dasd.dasdview')
    def test_get_dasd_info(self, m_dview):
        devname = random_string()
        id_path = random_string()
        device = {'DEVNAME': devname, 'ID_PATH': 'ccw-' + id_path}
        m_dview.return_value = self._load_test_data('dasdd.view')
        self.assertEqual(
            update_probe_data(
                expected_probe_data['/dev/dasdd'],
                name=devname, device_id=id_path),
            dasd.get_dasd_info(device))

    @mock.patch('probert.dasd.dasdview')
    def test_get_dasd_info_returns_none_if_not_all(self, m_dview):
        devname = random_string()
        id_path = random_string()
        device = {'DEVNAME': devname, 'ID_PATH': 'ccw-' + id_path}
        m_dview.return_value = random_string()
        self.assertIsNone(dasd.get_dasd_info(device))

    @mock.patch('probert.dasd.find_val_int')
    @mock.patch('probert.dasd.dasdview')
    def test_get_dasd_info_returns_none_if_bad_blocksize(self, m_dview,
                                                         m_find_val_int):
        devname = random_string()
        id_path = random_string()
        device = {'DEVNAME': devname, 'ID_PATH': 'ccw-' + id_path}
        m_dview.return_value = self._load_test_data('dasdd.view')
        m_find_val_int.return_value = None
        self.assertIsNone(dasd.get_dasd_info(device))

    @mock.patch('probert.dasd.disk_format')
    @mock.patch('probert.dasd.dasdview')
    def test_get_dasd_info_returns_none_if_bad_disk_format(self, m_dview,
                                                           m_disk):
        devname = random_string()
        id_path = random_string()
        device = {'DEVNAME': devname, 'ID_PATH': 'ccw-' + id_path}
        m_dview.return_value = self._load_test_data('dasdd.view')
        m_disk.return_value = None
        self.assertIsNone(dasd.get_dasd_info(device))

    @mock.patch('probert.dasd.platform.machine')
    def test_dasd_probe_returns_empty_dict_non_s390x_arch(self, m_machine):
        machine = random_string()
        self.assertNotEqual("s390x", machine)
        m_machine.return_value = machine
        self.assertEqual({}, dasd.probe())

    @mock.patch('probert.dasd.platform.machine')
    @mock.patch('probert.dasd.dasdview')
    def test_dasd_probe_dasdd(self, m_dasdview, m_machine):
        m_machine.return_value = 's390x'
        m_dasdview.side_effect = iter([self._load_test_data('dasdd.view')])

        context = mock.MagicMock()
        context.list_devices.side_effect = iter([
            [{"MAJOR": "94", "DEVNAME": "/dev/dasdd", "ID_SERIAL": "0X1544",
             "ID_PATH": "ccw-0.0.1544"}],
        ])
        expected_results = {
            '/dev/dasdd': update_probe_data(
                expected_probe_data['/dev/dasdd'], device_id="0.0.1544")
        }
        self.assertEqual(expected_results, dasd.probe(context=context))

    @mock.patch('probert.dasd.platform.machine')
    @mock.patch('probert.dasd.dasdview')
    def test_dasd_probe_dasde(self, m_dasdview, m_machine):
        m_machine.return_value = 's390x'
        m_dasdview.side_effect = iter([self._load_test_data('dasde.view')])

        context = mock.MagicMock()
        context.list_devices.side_effect = iter([
            [{"MAJOR": "94", "DEVNAME": "/dev/dasde",
             "ID_PATH": "ccw-0.0.2250"}],
        ])
        expected_results = {
            '/dev/dasde': update_probe_data(
                expected_probe_data['/dev/dasde'], device_id="0.0.2250")
            }
        self.assertEqual(expected_results, dasd.probe(context=context))

    @mock.patch('probert.dasd.platform.machine')
    @mock.patch('probert.dasd.dasdview')
    def test_dasd_probe_dasdd_skips_partitions(self, m_dasdview, m_machine):
        m_machine.return_value = 's390x'
        m_dasdview.side_effect = iter([self._load_test_data('dasdd.view')])

        context = mock.MagicMock()
        context.list_devices.side_effect = iter([
            [{"MAJOR": "94", "DEVNAME": "/dev/dasdd", "ID_SERIAL": "0X1544",
             "ID_PATH": "ccw-0.0.1544"}],
            [{"MAJOR": "94", "DEVNAME": "/dev/dasdd1", "ID_SERIAL": "0X1544",
             "ID_PATH": "ccw-0.0.1544", "PARTN": "1"}],
        ])
        expected_results = {
            '/dev/dasdd': update_probe_data(
                expected_probe_data['/dev/dasdd'], device_id="0.0.1544"),
            }
        self.assertEqual(expected_results, dasd.probe(context=context))

    @mock.patch('probert.dasd.subprocess.run')
    @mock.patch('probert.dasd.open')
    @mock.patch('probert.dasd.platform.machine')
    def test_dasd_probe_virtio_dasd(self, m_machine, m_open, m_run):
        m_machine.return_value = 's390x'

        virtio_major = random_string()
        devname = random_string()

        m_open.return_value = ['{} virtblk\n'.format(virtio_major)]
        m_run.return_value.returncode = 0

        context = mock.MagicMock()
        context.list_devices.side_effect = iter([
            [{"MAJOR": virtio_major, "DEVNAME": devname}],
        ])
        expected_results = {
            devname: {'name': devname, 'type': 'virt'}
            }
        self.assertEqual(expected_results, dasd.probe(context=context))
        m_run.assert_called_once_with(
            ['fdasd', '-i', devname],
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    @mock.patch('probert.dasd.subprocess.run')
    @mock.patch('probert.dasd.open')
    @mock.patch('probert.dasd.platform.machine')
    def test_dasd_probe_virtio_non_dasd(self, m_machine, m_open, m_run):
        m_machine.return_value = 's390x'

        virtio_major = random_string()
        devname = random_string()

        m_open.return_value = ['{} virtblk\n'.format(virtio_major)]
        m_run.return_value.returncode = 1

        context = mock.MagicMock()
        context.list_devices.side_effect = iter([
            [{"MAJOR": virtio_major, "DEVNAME": devname}],
        ])
        self.assertEqual({}, dasd.probe(context=context))
        m_run.assert_called_once_with(
            ['fdasd', '-i', devname],
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import subprocess

from probert.utils import sane_block_devices


log = logging.getLogger('probert.bcache')


def superblock_asdict(device=None, data=None):
    """ Convert output from bcache-super-show into a dictionary"""

    if not device and not data:
        raise ValueError('Supply a device name, or data to parse')

    if not data:
        cmd = ['bcache-super-show', device]
        result = subprocess.run(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.DEVNULL)
        data = result.stdout.decode('utf-8')
    bcache_super = {}
    for line in data.splitlines():
        if not line:
            continue
        values = [val for val in line.split('\t') if val]
        if len(values) == 2:
            bcache_super.update({values[0]: values[1]})

    return bcache_super


def parse_sb_version(sb_version):
    """ Convert sb_version string to integer if possible"""
    try:
        # 'sb.version': '1 [backing device]'
        # 'sb.version': '3 [caching device]'
        version = int(sb_version.split()[0])
    except (AttributeError, ValueError):
        log.warning("Failed to parse bcache 'sb.version' field"
                    " as integer: %s", sb_version)
        return None

    return version


def is_backing(device):
    """ Test if device is a bcache backing device

    A runtime check for an active bcache backing device is to
    examine /sys/class/block/<kname>/bcache/label

    However if a device is not active then read the superblock
    of the device and check that sb.version == 1"""

    sys_block = '/sys/class/block/%s' % os.path.basename(device)
    bcache_sys_attr = os.path.join(sys_block, 'bcache', 'label')
    return os.path.exists(bcache_sys_attr)


def is_caching(device):
    """ Test if device is a bcache caching device

    A runtime check for an active bcache backing device is to
    examine /sys/class/block/<kname>/bcache/cache_replacement_policy

    However if a device is not active then read the superblock
    of the device and check that sb.version == 3"""

    sys_block = '/sys/class/block/%s' % os.path.basename(device)
    bcache_sys_attr = os.path.join(sys_block, 'bcache',
                                   'cache_replacement_policy')
    return os.path.exists(bcache_sys_attr)


def is_bcache_device(device):
    return device.get('ID_FS_TYPE') == 'bcache'


def probe(context=None):
    """Probe the system for bcache devices.  Bcache devices
       are registered with the kernel upon module load and when
       devices are hot/cold plugged.  There are two portions to
       a bcache, the backing device which holds data and the cache
       device which cached data from the backing device.  A backing
       device encodes a specific cache_set UUID in the backing device
       which is used to bind both device in the kernel and create a
       new block device, bcacheN.

       For each block device which has a bcache superblock embedded,
       extract and examine the superblock to determine which type of
       bcache device (backing, caching) and the relevant UUIDs and
       build (if possible) the pairing of caches to backing.

       This probe reports the devices separately but enough information
       is included to re-assemble joined bcache devices if desired.
    """
    backing = {}
    caching = {}
    bcache = {'backing': backing, 'caching': caching}
    if not context:
        return bcache

    for device in sane_block_devices(context):
        if is_bcache_device(device):
            devpath = device['DEVNAME']
            sb = superblock_asdict(devpath)
            bkey = sb.get('dev.uuid', 'not available')
            bconfig = {'blockdev': devpath, 'superblock': sb}
            if is_backing(devpath):
                backing[bkey] = bconfig
            elif is_caching(devpath):
                caching[bkey] = bconfig
            else:
                log.error('bcache.probe: %s is not bcache' % devpath)

    return bcache



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import subprocess

log = logging.getLogger('probert.mount')


def findmnt(data=None):
    if not data:
        cmd = ['findmnt', '--bytes', '--json']
        try:
            result = subprocess.run(cmd, stdout=subprocess.PIPE,
                                    stderr=subprocess.DEVNULL)
        except (subprocess.CalledProcessError, FileNotFoundError):
            return {}

        data = result.stdout.decode('utf-8')

    mounts = {}
    try:
        mounts = json.loads(data)
    except json.decoder.JSONDecodeError as e:
        log.error('Failed to load findmnt json output:', e)

    return mounts


def probe(context=None):
    """The probert uses the util-linux 'findmnt' command which
       dumps a JSON tree of detailed information about _all_
       mounts in the current linux system.
    """
    return findmnt().get('filesystems', {})



# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class Prober():
    def __init__(self):
        self._results = {}

    def probe_all(self):
        self.probe_storage()
        self.probe_network()

    def probe_storage(self):
        from probert.storage import Storage
        self._storage = Storage()
        self._results['storage'] = self._storage.probe()

    def probe_network(self):
        from probert.network import NetworkProber
        self._network = NetworkProber()
        self._results['network'] = self._network.probe()

    def get_results(self):
        return self._results



import testtools
import json

from probert.storage import Storage, StorageInfo
from probert.tests.fakes import FAKE_PROBE_ALL_JSON


class ProbertTestStorage(testtools.TestCase):
    def setUp(self):
        super(ProbertTestStorage, self).setUp()
        with open(FAKE_PROBE_ALL_JSON) as f:
            self.results = json.load(f)
        self.storage = Storage(results=self.results)

    def test_storage_init(self):
        self.assertNotEqual(None, self.storage)


class ProbertTestStorageInfo(testtools.TestCase):
    ''' properties:
        .name = /dev/sda
        .type = disk
        .vendor = SanDisk
        .model = SanDisk_12123123
        .serial = aaccasdf
        .devpath = /devices
        .is_virtual =
        .raw = {raw dictionary}
    '''
    def setUp(self):
        super(ProbertTestStorageInfo, self).setUp()
        with open(FAKE_PROBE_ALL_JSON) as f:
            self.results = json.load(f)

    def test_storageinfo_init(self):
        probe_data = {
            '/dev/sda': {
                'DEVTYPE': 'disk',
                'attrs': {
                    'size': '1000000'
                }
            }
        }
        si = StorageInfo(probe_data)
        self.assertNotEqual(si, None)

    def test_storageinfo_attributes(self):
        sda = {'/dev/sda': self.results.get('storage').get('/dev/sda')}
        si = StorageInfo(probe_data=sda)
        props = {
            'name': '/dev/sda',
            'type': 'disk',
            'vendor': 'SanDisk',
            'model': 'SanDisk_SD5SG2128G1052E',
            'serial': 'SanDisk_SD5SG2128G1052E_133507400177',
            'devpath': (
                '/devices/pci0000:00/0000:00:1f.2/ata1/'
                'host0/target0:0:0/0:0:0:0/block/sda'
            ),
            'is_virtual': False,
            'raw': sda.get('/dev/sda')
        }
        for (prop, value) in props.items():
            self.assertEqual(getattr(si, prop), value)



import testtools
from mock import call

from probert import utils
from probert.tests.helpers import random_string, simple_mocked_open


class ProbertTestUtils(testtools.TestCase):
    def setUp(self):
        super(ProbertTestUtils, self).setUp()

    def test_utils_dict_merge(self):
        r1 = {'relations': ['m1', 'x1']}
        r2 = {'relations': ['m2', 'x2']}
        combined = {'relations': ['m1', 'm2', 'x1', 'x2']}
        test_result = utils.dict_merge(r1, r2)
        self.assertEqual(sorted(combined['relations']),
                         sorted(test_result['relations']))

    def test_utils_dict_merge_lists(self):
        r1 = ['m1', 'x1']
        r2 = ['m2', 'x2']
        combined = ['m1', 'm2', 'x1', 'x2']
        test_result = utils.dict_merge(r1, r2)
        self.assertEqual(sorted(combined), sorted(test_result))

    def test_utils_dict_merge_dicts(self):
        r1 = {'storage': {'/dev/sda': {'DEVTYPE': 'disk'}}}
        r2 = {'storage': {'/dev/sda': {'ID_MODEL': 'AWESOME'}}}
        combined = {
            'storage': {
                '/dev/sda': {
                    'DEVTYPE': 'disk',
                    'ID_MODEL': 'AWESOME',
                }
            }
        }
        test_result = utils.dict_merge(r1, r2)
        self.assertEqual(sorted(combined), sorted(test_result))

    def test_utils_read_sys_block_size_bytes(self):
        devname = random_string()
        expected_fname = '/sys/class/block/%s/size' % devname
        expected_bytes = 10737418240
        content = '20971520'
        with simple_mocked_open(content=content) as m_open:
            result = utils.read_sys_block_size_bytes(devname)
            self.assertEqual(expected_bytes, result)
            self.assertEqual([call(expected_fname)], m_open.call_args_list)

    def test_utils_read_sys_block_size_bytes_strips_value(self):
        devname = random_string()
        expected_fname = '/sys/class/block/%s/size' % devname
        expected_bytes = 10737418240
        content = ' 20971520 \n '
        with simple_mocked_open(content=content) as m_open:
            result = utils.read_sys_block_size_bytes(devname)
            self.assertEqual(expected_bytes, result)
            self.assertEqual([call(expected_fname)], m_open.call_args_list)



"""C extension build configuration for probert (libnl bindings)."""

import subprocess

from setuptools import Extension, setup


def pkgconfig(package):
    return {
        "extra_compile_args": subprocess.check_output(
            ["pkg-config", "--cflags", package]
        ).decode("utf8").split(),
        "extra_link_args": subprocess.check_output(
            ["pkg-config", "--libs", package]
        ).decode("utf8").split(),
    }


setup(
    ext_modules=[
        Extension(
            "probert._rtnetlink",
            ["probert/_rtnetlinkmodule.c"],
            **pkgconfig("libnl-route-3.0"),
        ),
        Extension(
            "probert._nl80211",
            ["probert/_nl80211module.c"],
            **pkgconfig("libnl-genl-3.0"),
        ),
    ],
)



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

import pyudev

from probert.utils import sane_block_devices

log = logging.getLogger('probert.filesystems')


def get_device_filesystem(device):
    # extract ID_FS_* keys into dict, dropping leading ID_FS
    return {k.replace('ID_FS_', ''): v
            for k, v in device.items() if k.startswith('ID_FS_')}


def probe(context=None):
    """ Capture detected filesystems found on discovered block devices.  """
    filesystems = {}
    if not context:
        context = pyudev.Context()

    for device in sane_block_devices(context):
        # Ignore block major=1 (ramdisk) and major=7 (loopback)
        # these won't ever be used in recreating storage on target systems.
        if device['MAJOR'] not in ["1", "7"]:
            fs_info = get_device_filesystem(device)
            # The ID_FS_ udev values come from libblkid, which contains code to
            # recognize lots of different things that block devices or their
            # partitions can contain (filesystems, lvm PVs, bcache, ...).  We
            # only want to report things that are mountable filesystems here,
            # which libblkid conveniently tags with ID_FS_USAGE=filesystem.
            # Swap is a bit of a special case because it is not a mountable
            # filesystem in the usual sense, but subiquity still needs to
            # generate mount actions for it.  Crypto is a disguised filesystem.
            if fs_info.get("USAGE") in ("filesystem", "crypto") or \
               fs_info.get("TYPE") == "swap":
                filesystems[device['DEVNAME']] = fs_info

    return filesystems



# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from collections import OrderedDict
import contextlib
import ipaddress
import jsonschema
import logging
import os
import socket

import pyudev

from probert.utils import udev_get_attributes

log = logging.getLogger('probert.network')

try:
    from probert import _nl80211, _rtnetlink
except ImportError as e:
    log.warning('Failed import network library modules: %s', e)

# Standard interface flags (net/if.h)
IFF_UP = 0x1                   # Interface is up.
IFF_BROADCAST = 0x2            # Broadcast address valid.
IFF_DEBUG = 0x4                # Turn on debugging.
IFF_LOOPBACK = 0x8             # Is a loopback net.
IFF_POINTOPOINT = 0x10         # Interface is point-to-point link.
IFF_NOTRAILERS = 0x20          # Avoid use of trailers.
IFF_RUNNING = 0x40             # Resources allocated.
IFF_NOARP = 0x80               # No address resolution protocol.
IFF_PROMISC = 0x100            # Receive all packets.
IFF_ALLMULTI = 0x200           # Receive all multicast packets.
IFF_MASTER = 0x400             # Master of a load balancer.
IFF_SLAVE = 0x800              # Slave of a load balancer.
IFF_MULTICAST = 0x1000         # Supports multicast.
IFF_PORTSEL = 0x2000           # Can set media type.
IFF_AUTOMEDIA = 0x4000         # Auto media select active.

IFA_F_PERMANENT = 0x80

BOND_MODES = [
    "balance-rr",
    "active-backup",
    "balance-xor",
    "broadcast",
    "802.3ad",
    "balance-tlb",
    "balance-alb",
]

XMIT_HASH_POLICIES = [
    "layer2",
    "layer2+3",
    "layer3+4",
    "encap2+3",
    "encap3+4",
]

LACP_RATES = [
    "slow",
    "fast",
]

# This json schema describes the links as they are serialized onto
# disk by probert --network. It also describes the format of some of
# the attributes of Link instances.
link_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "title": "link",
    "type": "object",
    "additionalProperties": False,
    "required": ["addresses", "bond", "bridge", "netlink_data", "type",
                 "udev_data"],
    "properties": {
        "addresses": {
            "type": "array",
            "items": {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "address": {"type": "string"},
                    "ip": {"type": "string"},
                    "family": {"type": "integer"},
                    "source": {"type": "string"},
                    "scope": {"type": "string"},
                },
            },
        },
        "type": {
            "type": "string",
            # "enum": ["eth", "wlan", "bridge", "vlan"], # there are more
        },
        "bond": {
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "is_master": {"type": "boolean"},
                "is_slave": {"type": "boolean"},
                "master": {
                    "oneOf": [
                        {"type": "string"},
                        {"type": "null"},
                    ],
                },
                "slaves": {
                    "type": "array",
                    "items": {"type": "string"},
                },
                "mode": {
                    "oneOf": [
                        {"type": "string", "enum": BOND_MODES},
                        {"type": "null"},
                    ],
                },
                "xmit_hash_policy": {
                    "oneOf": [
                        {"type": "string", "enum": XMIT_HASH_POLICIES},
                        {"type": "null"},
                    ],
                },
                "lacp_rate": {
                    "oneOf": [
                        {"type": "string", "enum": LACP_RATES},
                        {"type": "null"},
                    ],
                },
            },
        },
        "udev_data": {
            "type": "object",
            "properties": {
                "attrs": {
                    "type": "object",
                    "additionalProperties": {
                        "oneOf": [
                            {"type": "string"},
                            {"type": "null"},
                        ],
                    },
                },
            },
            "additionalProperties": {
                "oneOf": [
                    {"type": "string"},
                    {"type": "null"},
                ],
            },
        },
        "netlink_data": {
            "type": "object",
            "properties": {
                "ifindex": {"type": "integer"},
                "flags": {"type": "integer"},
                "arptype": {"type": "integer"},
                "family": {"type": "integer"},
                "name": {"type": "string"},
            },
        },
        "bridge": {
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "is_bridge": {"type": "boolean"},
                "is_port": {"type": "boolean"},
                "interfaces": {"type": "array", "items": {"type": "string"}},
                "options": {  # /sys/class/net/brX/bridge/<options key>
                    "type": "object",
                    "additionalProperties": {"type": "string"},
                },
            },
        },
        "wlan": {
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "ssid": {"type": ["null", "string"]},
                "visible_ssids": {
                    "type": "array",
                    "items": {"type": "string"},
                },
                "scan_state": {"type": ["null", "string"]},
            },
        },
    },
}


def _compute_type(iface, arptype):
    if not iface:
        return '???'

    sysfs_path = os.path.join('/sys/class/net', iface)
    if not os.path.exists(sysfs_path):
        log.debug('No sysfs path to {}'.format(sysfs_path))
        return None

    DEV_TYPE = '???'

    if arptype == 1:
        DEV_TYPE = 'eth'
        if os.path.isdir(os.path.join(sysfs_path, 'wireless')) or \
           os.path.islink(os.path.join(sysfs_path, 'phy80211')):
            DEV_TYPE = 'wlan'
        elif os.path.isdir(os.path.join(sysfs_path, 'bridge')):
            DEV_TYPE = 'bridge'
        elif os.path.isdir(os.path.join(sysfs_path, 'bonding')):
            DEV_TYPE = 'bond'
        elif os.path.isfile(os.path.join(sysfs_path, 'tun_flags')):
            DEV_TYPE = 'tap'
        elif os.path.isdir(
                os.path.join('/sys/devices/virtual/net', iface)):
            if iface.startswith('dummy'):
                DEV_TYPE = 'dummy'
    elif arptype == 24:  # firewire ;; IEEE 1394 - RFC 2734
        DEV_TYPE = 'eth'
    elif arptype == 32:  # InfiniBand
        if os.path.isdir(os.path.join(sysfs_path, 'bonding')):
            DEV_TYPE = 'bond'
        elif os.path.isdir(os.path.join(sysfs_path, 'create_child')):
            DEV_TYPE = 'ib'
        else:
            DEV_TYPE = 'ibchild'
    elif arptype == 280:
        DEV_TYPE = 'can'
    elif arptype == 512:
        DEV_TYPE = 'ppp'
    elif arptype == 768:
        DEV_TYPE = 'ipip'      # IPIP tunnel
    elif arptype == 769:
        DEV_TYPE = 'ip6tnl'   # IP6IP6 tunnel
    elif arptype == 772:
        DEV_TYPE = 'lo'
    elif arptype == 776:
        DEV_TYPE = 'sit'      # sit0 device - IPv6-in-IPv4
    elif arptype == 778:
        DEV_TYPE = 'gre'      # GRE over IP
    elif arptype == 783:
        DEV_TYPE = 'irda'     # Linux-IrDA
    elif arptype == 801:
        DEV_TYPE = 'wlan_aux'
    elif arptype == 65534:
        DEV_TYPE = 'tun'

    if iface.startswith('ippp') or iface.startswith('isdn'):
        DEV_TYPE = 'isdn'
    elif iface.startswith('mip6mnha'):
        DEV_TYPE = 'mip6mnha'

    if len(DEV_TYPE) == 0:
        print('Failed to determine interface type for {}'.format(iface))
        return None

    return DEV_TYPE


def _get_bonding(ifname, flags):

    def _iface_is_master():
        return bool(flags & IFF_MASTER) != 0

    def _iface_is_slave():
        return bool(flags & IFF_SLAVE) != 0

    def _get_slave_iface_list():
        try:
            if _iface_is_master():
                bond = open('/sys/class/net/%s/bonding/slaves' % ifname).read()
                return bond.split()
            else:
                return []
        except IOError:
            return []

    def _get_bond_master():
        try:
            if _iface_is_slave():
                master = os.readlink('/sys/class/net/%s/master' % ifname)
                return os.path.basename(master)
            else:
                return None
        except IOError:
            return None

    def _get_bond_param(param):
        try:
            if _iface_is_master():
                bond_param = '/sys/class/net/%s/bonding/%s' % (ifname, param)
                with open(bond_param) as bp:
                    bond_param = bp.read().split()
                return bond_param[0] if bond_param else None
        except IOError:
            return None

    return {
        'is_master': _iface_is_master(),
        'is_slave': _iface_is_slave(),
        'master': _get_bond_master(),
        'slaves': _get_slave_iface_list(),
        'mode': _get_bond_param('mode'),
        'xmit_hash_policy': _get_bond_param('xmit_hash_policy'),
        'lacp_rate': _get_bond_param('lacp_rate'),
    }


def _get_bridging(ifname):

    def _iface_is_bridge():
        bridge_path = os.path.join('/sys/class/net', ifname, 'bridge')
        return os.path.exists(bridge_path)

    def _iface_is_bridge_port():
        bridge_port = os.path.join('/sys/class/net', ifname, 'brport')
        return os.path.exists(bridge_port)

    def _get_bridge_iface_list():
        if _iface_is_bridge():
            bridge_path = os.path.join('/sys/class/net', ifname, 'brif')
            return os.listdir(bridge_path)
        return []

    def _get_bridge_options():
        skip_attrs = set(['flush', 'bridge'])  # needs root access, not useful

        if _iface_is_bridge():
            bridge_path = os.path.join('/sys/class/net', ifname, 'bridge')
        elif _iface_is_bridge_port():
            bridge_path = os.path.join('/sys/class/net', ifname, 'brport')
        else:
            return {}

        options = {}
        for bridge_attr_name in os.listdir(bridge_path):
            if bridge_attr_name in skip_attrs:
                continue
            bridge_attr_file = os.path.join(bridge_path, bridge_attr_name)
            with open(bridge_attr_file) as bridge_attr:
                options[bridge_attr_name] = bridge_attr.read().strip()

        return options

    return {
        'is_bridge': _iface_is_bridge(),
        'is_port': _iface_is_bridge_port(),
        'interfaces': _get_bridge_iface_list(),
        'options': _get_bridge_options(),
    }


def netlink_attr(attr):
    def get(obj):
        return obj.netlink_data[attr]
    return property(get)


def udev_attr(keys, missing):
    def get(obj):
        for k in keys:
            if k in obj.udev_data:
                return obj.udev_data[k]
        return missing
    return property(get)


class Link:

    @classmethod
    def from_probe_data(cls, netlink_data, udev_data):
        # This is a bit of a hack, but sometimes the interface has
        # already been renamed by udev by the time we get here, so we
        # can't use netlink_data['name'] to go poking about in
        # /sys/class/net.
        name = socket.if_indextoname(netlink_data['ifindex'])
        if netlink_data['is_vlan']:
            typ = 'vlan'
        else:
            typ = _compute_type(name, netlink_data['arptype'])
        link = cls(
            addresses={},
            type=typ,
            udev_data=udev_data,
            netlink_data=netlink_data,
            bond=_get_bonding(name, netlink_data['flags']),
            bridge=_get_bridging(name))
        if udev_data.get('DEVTYPE') == 'wlan':
            link.wlan = {
                'visible_ssids': [],
                'ssid': None,
                'scan_state': None,
            }
        return link

    @classmethod
    def from_saved_data(cls, link_data):
        address_objs = {}
        for addr in link_data['addresses']:
            a = Address.from_saved_data(addr)
            address_objs[str(a.ip)] = a
        link_data['addresses'] = address_objs
        return cls(**link_data)

    def __init__(self, addresses, type, udev_data, netlink_data, bond,
                 bridge, wlan=None):
        self.addresses = addresses
        self.type = type
        self.udev_data = udev_data
        self.netlink_data = netlink_data
        self.bond = bond
        self.bridge = bridge
        self.wlan = wlan

    def serialize(self):
        r = {
            "addresses": [a.serialize() for a in self.addresses.values()],
            "udev_data": self.udev_data,
            "type": self.type,
            "netlink_data": self.netlink_data,
            "bond": self.bond,
            "bridge": self.bridge,
        }
        if self.wlan is not None:
            r["wlan"] = self.wlan
        jsonschema.validate(r, link_schema)
        return r

    flags = netlink_attr("flags")
    ifindex = netlink_attr("ifindex")
    name = netlink_attr("name")
    hwaddr = property(lambda self: self.udev_data['attrs']['address'])

    vendor = udev_attr(['ID_VENDOR_FROM_DATABASE', 'ID_VENDOR',
                        'ID_VENDOR_ID'], "Unknown Vendor")
    model = udev_attr(['ID_MODEL_FROM_DATABASE', 'ID_MODEL', 'ID_MODEL_ID'],
                      "Unknown Model")
    driver = udev_attr(['ID_NET_DRIVER', 'ID_USB_DRIVER'], "Unknown Driver")
    devpath = udev_attr(['DEVPATH'], "Unknown devpath")

    hwaddr = property(lambda self: self.udev_data['attrs']['address'])

    # This is the logic ip from iproute2 uses to determine whether
    # to show NO-CARRIER or not. It only really makes sense for a
    # wired connection.
    is_connected = (
        property(lambda self: (
            (not (self.flags & IFF_UP)) or (self.flags & IFF_RUNNING))))
    is_virtual = (
        property(lambda self: self.devpath.startswith('/devices/virtual/')))

    @property
    def ssid(self):
        if self.wlan:
            return self.wlan['ssid']
        else:
            return None


_scope_str = {
    0: 'global',
    200: "site",
    253: "link",
    254: "host",
    255: "nowhere",
}


class Address:

    def __init__(self, address, family, source, scope):
        self.address = ipaddress.ip_interface(address)
        self.ip = self.address.ip
        self.family = family
        self.source = source
        self.scope = scope

    def serialize(self):
        return {
            'source': self.source,
            'family': self.family,
            'address': str(self.address),
            'scope': self.scope,
        }

    @classmethod
    def from_probe_data(cls, netlink_data):
        address = netlink_data['local'].decode('latin-1')
        family = netlink_data['family']
        if netlink_data.get('flags', 0) & IFA_F_PERMANENT:
            source = 'static'
        else:
            source = 'dhcp'
        scope = netlink_data['scope']
        scope = str(_scope_str.get(scope, scope))
        return cls(address, family, source, scope)

    @classmethod
    def from_saved_data(cls, link_data):
        return Address(**link_data)


class NetworkObserver(abc.ABC):
    """A NetworkObserver observes the network state.

    It calls methods on a NetworkEventReceiver in response to changes.
    """

    @abc.abstractmethod
    def start(self):
        pass

    @abc.abstractmethod
    def data_ready(self, fd):
        pass


class NetworkEventReceiver(abc.ABC):
    """NetworkEventReceiver has methods called on it in response to network
       changes."""

    @abc.abstractmethod
    def new_link(self, ifindex, link):
        pass

    @abc.abstractmethod
    def update_link(self, ifindex):
        pass

    @abc.abstractmethod
    def del_link(self, ifindex):
        pass

    @abc.abstractmethod
    def route_change(self, action, data):
        pass


class TrivialEventReceiver(NetworkEventReceiver):

    def new_link(self, ifindex, link):
        pass

    def update_link(self, ifindex):
        pass

    def del_link(self, ifindex):
        pass

    def route_change(self, action, data):
        pass

# Coalescing netlink events
#
# If the client of this library delays calling UdevObserver.data_ready
# until the udev queue is idle (which is a good idea, but cannot be
# implemented here because delaying inherently depends on the event
# loop the client is using), several netlink events might be seen for
# any interface -- the poster child for this being when an interface
# is renamed by a udev rule. The netlink data that comes with the NEW
# event in this case can be out of date by the time the event is
# processed, so what the @coalesce generator does is to collapse a
# series of calls for one object into one, e.g. NEW + CHANGE becomes
# NEW but with the data from the change event, NEW + DEL is dropped
# entirely, etc. @nocoalesce doesn't combine any events but makes sure
# that those events are not processed wildly out of order with the
# events that are coalesced.


def coalesce(*keys):
    # "keys" defines which events are coalesced, ifindex is enough for
    # link events but ifindex + address is needed for address events.
    def decorator(func):
        def w(self, action, data):
            log.debug('event for %s: %s %s', func.__name__, action, data)
            key = (func.__name__,)
            for k in keys:
                key += (data[k],)
            if key in self._calls:
                prev_meth, prev_action, prev_data = self._calls[key]
                if action == 'NEW':
                    # this clearly shouldn't happen, but take the new data
                    # just in case
                    self._calls[key] = (func, action, data)
                elif action == 'CHANGE':
                    # Merge change data into previous data so attributes
                    # from the original NEW event (e.g. is_vlan, vlan_id,
                    # vlan_link) are preserved when a CHANGE event arrives
                    # without them.
                    merged = {**prev_data, **data}
                    self._calls[key] = (func, prev_action, merged)
                elif action == 'DEL':
                    if prev_action == 'NEW':
                        # link disappeared before we did anything with it.
                        # forget about it.
                        del self._calls[key]
                    else:
                        # Otherwise just pass on the DEL and forget the
                        # previous action whatever it was.
                        self._calls[key] = (func, action, data)
            else:
                self._calls[key] = (func, action, data)
        return w
    return decorator


def nocoalesce(func):
    def w(self, action, data):
        self._calls[object()] = (func, action, data)
    return w


@contextlib.contextmanager
def CoalescedCalls(obj):
    obj._calls = OrderedDict()
    try:
        yield
    finally:
        for meth, action, data in obj._calls.values():
            meth(obj, action, data)
        obj._calls = None


class UdevObserver(NetworkObserver):
    """Use udev/netlink to observe network changes."""

    def __init__(self, receiver=None):
        self._links = {}
        self.context = pyudev.Context()
        if receiver is None:
            receiver = TrivialEventReceiver()
        assert isinstance(receiver, NetworkEventReceiver)
        self.receiver = receiver
        self._calls = None

    def start(self):
        self.rtlistener = _rtnetlink.listener(self)
        with CoalescedCalls(self):
            self.rtlistener.start()

        self._fdmap = {
            self.rtlistener.fileno(): self.rtlistener.data_ready,
        }

        try:
            self.wlan_listener = _nl80211.listener(self)
            self.wlan_listener.start()
            self._fdmap.update({
                self.wlan_listener.fileno(): self.wlan_listener.data_ready,
            })
        except RuntimeError:
            log.debug('could not start wlan_listener')

        return list(self._fdmap)

    def data_ready(self, fd):
        with CoalescedCalls(self):
            self._fdmap[fd]()

    @coalesce('ifindex')
    def link_change(self, action, data):
        log.debug('link_change %s %s', action, data)
        for k, v in data.items():
            if isinstance(v, bytes):
                data[k] = v.decode('utf-8', 'replace')
        ifindex = data['ifindex']
        if action == 'DEL':
            if ifindex in self._links:
                del self._links[ifindex]
                self.receiver.del_link(ifindex)
            return
        if action == 'CHANGE':
            if ifindex in self._links:
                dev = self._links[ifindex]
                # Trigger a scan when a wlan device goes up
                # Not sure if this is required as devices seem to scan as soon
                # as they go up? (in which case this fails with EBUSY, so it's
                # just spam in the logs).
                if dev.type == 'wlan':
                    if (not (dev.flags & IFF_UP)) and (data['flags'] & IFF_UP):
                        try:
                            self.trigger_scan(ifindex)
                        except RuntimeError:
                            log.exception('on-up trigger_scan failed')
                dev.netlink_data = data
                # If a device appears and is immediately renamed, the
                # initial _compute_type can fail to find the sysfs
                # directory. Have another go now.
                if dev.type is None:
                    dev.type = _compute_type(dev.name)
                dev.bond = _get_bonding(dev.name, dev.netlink_data['flags'])
            self.receiver.update_link(ifindex)
            return
        udev_devices = list(self.context.list_devices(IFINDEX=str(ifindex)))
        if len(udev_devices) == 0:
            # Has disappeared already?
            return
        udev_device = udev_devices[0]
        udev_data = dict(udev_device)
        udev_data['attrs'] = udev_get_attributes(udev_device)
        link = Link.from_probe_data(data, udev_data)
        self._links[ifindex] = link
        self.receiver.new_link(ifindex, link)

    @coalesce('ifindex', 'local')
    def addr_change(self, action, data):
        log.debug('addr_change %s %s', action, data)
        link = self._links.get(data['ifindex'])
        if link is None:
            return
        ip = data['local'].decode('latin-1')
        if action == 'DEL':
            link.addresses.pop(ip, None)
            self.receiver.update_link(data['ifindex'])
            return
        link.addresses[ip] = Address.from_probe_data(data)
        self.receiver.update_link(data['ifindex'])

    @nocoalesce
    def route_change(self, action, data):
        log.debug('route_change %s %s', action, data)
        for k, v in data.items():
            if isinstance(v, bytes):
                data[k] = v.decode('utf-8', 'replace')
        self.receiver.route_change(action, data)

    def trigger_scan(self, ifindex):
        self.wlan_listener.trigger_scan(ifindex)

    def wlan_event(self, arg):
        log.debug('wlan_event %s', arg)
        ifindex = arg['ifindex']
        if ifindex < 0 or ifindex not in self._links:
            return
        link = self._links[ifindex]
        if arg['cmd'] == 'TRIGGER_SCAN':
            link.wlan['scan_state'] = 'scanning'
        if arg['cmd'] == 'NEW_SCAN_RESULTS' and 'ssids' in arg:
            ssids = set()
            for (ssid, status) in arg['ssids']:
                ssid = ssid.decode('utf-8', 'replace')
                ssids.add(ssid)
                if status != "no status":
                    link.wlan['ssid'] = ssid
            link.wlan['visible_ssids'] = sorted(ssids)
            link.wlan['scan_state'] = None
        if arg['cmd'] == 'NEW_INTERFACE':
            if link.flags & IFF_UP:
                try:
                    self.trigger_scan(ifindex)
                except RuntimeError:
                    # Can't trigger a scan as non-root, that's OK.
                    log.exception('initial trigger_scan failed')
            else:
                try:
                    self.rtlistener.set_link_flags(ifindex, IFF_UP)
                except RuntimeError:
                    log.exception('set_link_flags failed')
        if arg['cmd'] == 'NEW_INTERFACE' or arg['cmd'] == 'ASSOCIATE':
            if len(arg.get('ssids', [])) > 0:
                link.wlan['ssid'] = (
                    arg['ssids'][0][0].decode('utf-8', 'replace'))
        if arg['cmd'] == 'DISCONNECT':
            link.wlan['ssid'] = None


class StoredDataObserver:
    """A cheaty observer that just pretends the network is in some
       pre-arranged state."""

    def __init__(self, saved_data, receiver):
        self.saved_data = saved_data
        for data in self.saved_data['links']:
            jsonschema.validate(data, link_schema)
        self.links = {}
        self.receiver = receiver
        self.rd, self.wr = os.pipe()

    def start(self):
        for data in self.saved_data['links']:
            link = Link.from_saved_data(data)
            self.receiver.new_link(link.ifindex, link)
            self.links[link.ifindex] = link
        for data in self.saved_data['routes']:
            self.receiver.route_change("NEW", data)
        return [self.rd]

    def _scan_results(self, link):
        link.wlan['visible_ssids'] = ["AA", "BB"]
        link.wlan['scan_state'] = None
        os.write(self.wr, b'x')

    def trigger_scan(self, ifindex):
        import asyncio
        link = self.links[ifindex]
        link.wlan['scan_state'] = 'scanning'
        asyncio.get_event_loop().call_later(
            2.0, self._scan_results, link)
        os.write(self.wr, b'x')

    def data_ready(self, fd):
        os.read(self.rd, 1)


class NetworkProber:

    def probe(self):
        class CollectingReceiver(TrivialEventReceiver):
            def __init__(self):
                self.all_links = set()
                self.route_data = []

            def new_link(self, ifindex, link):
                self.all_links.add(link)

            def route_change(self, action, data):
                self.route_data.append(data)
        collector = CollectingReceiver()
        observer = UdevObserver(collector)
        observer.start()
        results = {
            'links': [],
            'routes': [],
        }
        for link in collector.all_links:
            results['links'].append(link.serialize())
        for route_data in collector.route_data:
            results['routes'].append(route_data)
        return results


if __name__ == '__main__':
    import pprint
    import select
    c = UdevObserver()
    fds = c.start()

    pprint.pprint(c.links)

    poll_ob = select.epoll()
    for fd in fds:
        poll_ob.register(fd, select.EPOLLIN)
    while True:
        events = poll_ob.poll()
        for (fd, e) in events:
            c.data_ready(fd)
        pprint.pprint(c.links)



import testtools
from unittest.mock import patch

from probert.prober import Prober
from probert.storage import Storage
from probert.network import NetworkProber


class ProbertTestProber(testtools.TestCase):

    def test_prober_init(self):
        p = Prober()
        self.assertNotEqual(p, None)

    @patch.object(Prober, 'probe_all')
    def test_prober_probe_all(self, _probe_all):
        p = Prober()
        p.probe_all()
        self.assertTrue(_probe_all.called)

    @patch.object(Prober, 'probe_network')
    @patch.object(Prober, 'probe_storage')
    def test_prober_probe_all_invoke_others(self, _storage, _network):
        p = Prober()
        p.probe_all()
        self.assertTrue(_storage.called)
        self.assertTrue(_network.called)

    def test_prober_get_results(self):
        p = Prober()
        self.assertEqual({}, p.get_results())

    @patch.object(NetworkProber, 'probe')
    @patch.object(Storage, 'probe')
    def test_prober_probe_all_check_results(self, _storage, _network):
        p = Prober()
        results = {
            'storage': {'lambic': 99},
            'network': {'saison': 99},
        }
        _storage.return_value = results['storage']
        _network.return_value = results['network']
        p.probe_all()
        self.assertTrue(_storage.called)
        self.assertTrue(_network.called)
        self.assertEqual(results, p.get_results())
